from typing import Any, Callable, Dict, List, Optional, Tuple, Union

from arkaine.internal.store.embeddings import (
    EmbeddingStore,
    InMemoryEmbeddingStore,
)
from arkaine.tools.tool import Argument, Context, Tool
from arkaine.tools.wrapper import Wrapper
from arkaine.utils.documents import chunk_text_by_sentences
from arkaine.utils.embeddings.model import OllamaEmbeddingModel


class TopN(Wrapper):
    """A wrapper tool that filters another tool's output using semantic search.

    This tool takes the output from another tool, chunks it into sentences, and
    returns the most semantically relevant sections based on a user query. It
    uses embeddings to perform the semantic search.

    If the tool being wrapped returns a string, it is simply chunked into
    sentences. If, however, the tool returns a list of strings, each item is
    individually chunked, maintaining separation of sections. If the output is
    a dictionary of [str, str], the .values() are utilized.

    Args:
        tool (Tool): The base tool to wrap and filter results from the wrapped
            tool
        n (int): Number of closest results to return
        embedder (Optional[InMemoryEmbeddingStore]): Custom embedding store to
            use. If None, a fresh InMemoryEmbeddingStore is created for each
            call. If a class is provided, it will be instantiated with
            embedder_kwargs for each call. If an object instance is provided,
            it will be reused for all calls (embedder_kwargs ignored).
        embedder_kwargs (Optional[Dict[str, Any]]): Arguments to initialize the
            embedder class. Required when embedder is a class. Ignored when
            embedder is None or an instance. Example: {"model":
            "text-embedding-3-small"}
        sentences_per (int, optional): Number of sentences per chunk. Defaults
            to 3.
        tool_formatter (Callable[[str], Union[str, List[str]]], optional):
            Custom formatter for the tool's output. This modifies the output to
            a string or a list of strings before being chunked. If not
            provided, we utilize the raw tool output.
        output_formatter (Callable[[str], str], optional): Custom formatter for
            the output of the top N results. This modifies the output to a
            string for future consumption. By default (when None), it will
            print "Here are the top {N} most relevant sections to the query:"
            followed by the results numerically ordered.
        name (str, optional): Custom name for the tool. Defaults to
            "{tool.name}::top_{n}".
        description (str, optional): Custom description. Defaults to base
            tool's description.
        query_attribute (str, optional): Name of the query parameter. Defaults
            to "query".
        query_description (str, optional): Description of the query parameter.
            Defaults to "The query to search for in the content".
    """

    def __init__(
        self,
        tool: Tool,
        n: int,
        embedder: Optional[EmbeddingStore] = None,
        embedder_kwargs: Optional[Dict[str, Any]] = None,
        sentences_per: int = 3,
        tool_formatter: Optional[Callable[[str], Union[str, List[str]]]] = None,
        output_formatter: Optional[Callable[[str], str]] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        query_attribute: str = "query",
        query_description: Optional[str] = None,
    ):
        # Validate embedder and embedder_kwargs combination
        if embedder is None:
            self._embedder = InMemoryEmbeddingStore
            self._embedder_kwargs = {"embedding_model": OllamaEmbeddingModel()}
        elif isinstance(embedder, type):
            if embedder_kwargs is None:
                raise ValueError(
                    "embedder_kwargs required when embedder is a class"
                )
            self._embedder = embedder
            self._embedder_kwargs = embedder_kwargs
        else:
            if embedder_kwargs:
                raise ValueError(
                    "embedder_kwargs should be None when embedder is an "
                    + "instance"
                )
            self._embedder = embedder
            self._embedder_kwargs = None

        self._n = n
        self._sentences_per = sentences_per
        self._tool_formatter = tool_formatter
        self._output_formatter = output_formatter
        self._query_attribute = query_attribute

        if not name:
            name = f"{tool.name}::top_{n}"

        if not description:
            description = tool.description

        if not query_description:
            query_description = "The query to search for in the content"

        args = [
            Argument(
                name=query_attribute,
                description=query_description,
                type="str",
                required=True,
            )
        ]

        super().__init__(
            name=name, description=description, tool=tool, args=args
        )

    def preprocess(self, ctx: Context, **kwargs) -> Tuple[Dict[str, Any], Any]:
        """Extract query and prepare arguments for the wrapped tool."""
        if self._query_attribute not in kwargs:
            raise ValueError(
                f"The {self._query_attribute} argument is required for this "
                + "tool"
            )

        # Extract query and pass it through to postprocess
        query = kwargs.pop(self._query_attribute)

        # Return processed args and the query as pass-through data
        return kwargs, query

    def postprocess(
        self, ctx: Context, passed: Optional[Any] = None, results: Any = None
    ) -> str:
        """Process the tool results using semantic search."""
        query = passed

        if self._tool_formatter:
            results = self._tool_formatter(results)

        if isinstance(results, str):
            results = [results]
        elif isinstance(results, dict):
            results = list(results.values())

        chunks = [
            chunk_text_by_sentences(item, self._sentences_per)
            for item in results
        ]

        # Initialize embedder based on configuration
        if self._embedder_kwargs is not None:
            embedder = self._embedder(**self._embedder_kwargs)
        else:
            embedder = self._embedder

        for chunk in chunks:
            embedder.add_text(chunk)

        search_results = embedder.query(query, top_n=self._n)

        if self._output_formatter:
            return self._output_formatter(search_results)

        out = (
            f"Here are the top {self._n} most relevant sections "
            + "to the query:\n"
        )

        for index, result in enumerate(search_results):
            out += f"{index + 1} - {result}\n"

        return out



import os
from abc import ABC, abstractmethod
from typing import List, Optional, Union

import ollama
import openai


class EmbeddingModel(ABC):

    def __init__(self):
        super().__init__()

    @abstractmethod
    def embed(self, text: Union[str, List[str]]) -> List[List[float]]:
        pass


class OllamaEmbeddingModel(EmbeddingModel):

    def __init__(self, model: str = "all-minilm:latest"):
        self.__embedding_model = model
        super().__init__()
        self.pull_model()

    def pull_model(self):
        ollama.pull(self.__embedding_model)

    def embed(self, text: Union[str, List[str]]) -> List[List[float]]:
        vectors: List[List[float]] = []

        if isinstance(text, str):
            text = [text]

        for t in text:
            vectors.append(
                ollama.embeddings(model=self.__embedding_model, prompt=t)[
                    "embedding"
                ]
            )

        return vectors


class OpenAIEmbeddingModel(EmbeddingModel):

    def __init__(
        self,
        model: str = "text-embedding-ada-002",
        api_key: Optional[str] = None,
    ):
        self.__embedding_model = model
        api_key = api_key or os.getenv("OPENAI_API_KEY")
        self.__client = openai.Client(api_key=api_key)

        super().__init__()

    def embed(self, text: Union[str, List[str]]) -> List[List[float]]:
        if isinstance(text, str):
            text = [text]

        response = self.__client.embeddings.create(
            input=text,
            model=self.__embedding_model,
        )
        return [data.embedding for data in response.data]



import json
from typing import Any, List, Optional


class Argument:
    def __init__(
        self,
        name: str,
        description: str,
        type: str,
        required: bool = False,
        default: Optional[Any] = None,
    ):
        self.name = name
        self.description = description
        self.type = type
        self.required = required
        self.default = (
            self._convert_value(default, type.lower()) if default else None
        )

    def _convert_value(self, value: Any, type_str: str) -> Any:
        """Convert a value to the appropriate type."""
        if value is None:
            return None

        if isinstance(value, str):
            # Handle basic types
            if type_str == "float":
                return float(value)
            elif type_str == "int":
                return int(value)
            elif type_str == "bool":
                return value.lower() == "true"
            # Handle lists and dicts
            elif type_str.startswith(("list", "dict")):
                try:
                    parsed = json.loads(value)
                    if type_str.startswith("list") and not isinstance(
                        parsed, list
                    ):
                        raise ValueError(f"Expected list, got {type(parsed)}")
                    if type_str.startswith("dict") and not isinstance(
                        parsed, dict
                    ):
                        raise ValueError(f"Expected dict, got {type(parsed)}")
                    return parsed
                except json.JSONDecodeError:
                    return value
        return value

    def __str__(self) -> str:
        out = f"{self.name} - {self.type_str} - Required: "
        out += f"{self.required} - "
        if self.default:
            out += f"Default: {self.default} - "
        out += f"{self.description}"

        return out

    def __repr__(self) -> str:
        return self.__str__()

    @property
    def type_str(self) -> str:
        """
        Since some might pass in the literal type instead of the str of the
        class, we should ensure we convert the type correctly to a string for
        parsing.

        It is not simply str(self.type) as that tends to add "<class 'type'>"
        to the string.
        """
        if isinstance(self.type, str):
            return self.type
        else:
            try:
                return str(self.type).split("'")[1]
            except Exception:
                return str(self.type)

    def to_json(self) -> dict:
        return {
            "name": self.name,
            "description": self.description,
            "type": self.type_str,
            "required": self.required,
            "default": self.default,
        }


class InvalidArgumentException(Exception):
    def __init__(
        self,
        tool_name: str,
        missing_required_args: List[str],
        extraneous_args: List[str],
    ):
        self.__tool_name = tool_name
        self.__missing_required_args = missing_required_args
        self.__extraneous_args = extraneous_args

    def __str__(self):
        out = f"Function {self.__tool_name} was improperly called\n"

        if self.__missing_required_args:
            out += (
                "Missing required arguments: "
                + ", ".join(self.__missing_required_args)
                + "\n"
            )
        if self.__extraneous_args:
            out += (
                "Extraneous arguments: "
                + ", ".join(self.__extraneous_args)
                + "\n"
            )

        return out



import os
from typing import Any, Callable, Dict, List, Optional, Union

import google.generativeai as genai
from google.ai.generativelanguage import GenerateContentResponse

from arkaine.backends.backend import Backend
from arkaine.backends.common import simple_tool_results_to_prompts
from arkaine.llms.google import Google as GoogleLLM
from arkaine.llms.llm import Prompt
from arkaine.tools.context import Context
from arkaine.tools.tool import Tool
from arkaine.tools.types import ToolCalls, ToolResults
from arkaine.utils.templater import PromptTemplate
from arkaine.utils.tool_format import gemini as gemini_format


class GoogleBackend(Backend):

    MODELS = GoogleLLM.MODELS

    def __init__(
        self,
        template: Union[PromptTemplate, str] = PromptTemplate.default(),
        model: str = "gemini-pro",
        tools: List[Tool] = [],
        initial_state: Dict[str, Any] = {},
        process_answer: Optional[Callable[[Any], Any]] = None,
        api_key: Optional[str] = None,
        context_length: Optional[int] = None,
    ):
        if api_key is None:
            api_key = os.environ.get("GOOGLE_AISTUDIO_API_KEY")
            if api_key is None:
                api_key = os.environ.get("GOOGLE_API_KEY")
            if api_key is None:
                raise ValueError(
                    "No Google API key found. Please set "
                    "GOOGLE_AISTUDIO_API_KEY or GOOGLE_API_KEY "
                    "environment variable"
                )

        genai.configure(api_key=api_key)
        self.__model = genai.GenerativeModel(model_name=model)

        if context_length:
            self.__context_length = context_length
        elif model in GoogleBackend.MODELS:
            self.__context_length = GoogleBackend.MODELS[model][
                "context_length"
            ]
        else:
            raise ValueError(
                f"Unknown model: {model} - must specify context length"
            )

        if isinstance(template, str):
            template = PromptTemplate(template)

        self.template = template

        super().__init__(
            None,
            tools,
            initial_state,
            process_answer,
        )

    def parse_for_tool_calls(
        self,
        context: Context,
        response: GenerateContentResponse,
        stop_at_first_tool: bool = False,
    ) -> ToolCalls:
        """Parse the response for tool calls.

        Args:
            context: The current execution context
            response: The model's response
            stop_at_first_tool: Whether to stop after finding the first tool
                call - ignored for this backend, included for abstract
                interface matching

        Returns:
            A list of (tool_name, tool_args) tuples
        """
        tool_calls = []

        # Extract function calls from response parts
        for candidate in response.candidates:
            for part in candidate.content.parts:
                if function_call := part.function_call:
                    # Convert MapComposite to regular dict
                    args_dict = dict(function_call.args)
                    tool_calls.append((function_call.name, args_dict))

        return tool_calls

    def parse_for_result(
        self, context: Context, response: GenerateContentResponse
    ) -> Optional[Any]:
        return response.candidates[0].content.parts[0].text

    def tool_results_to_prompts(
        self, context: Context, prompt: Prompt, results: ToolResults
    ) -> List[Prompt]:
        return simple_tool_results_to_prompts(prompt, results)

    def prepare_prompt(self, context: Context, **kwargs) -> Prompt:
        return self.template.render(kwargs)

    def query_model(
        self, context: Context, prompt: Prompt
    ) -> GenerateContentResponse:
        """Query the Gemini model with tools.

        Args:
            context: The current execution context
            prompt: The prompt to send to the model

        Returns:
            The model's response
        """
        tool_declarations = [
            gemini_format(tool) for tool in self.tools.values()
        ]

        history = []
        for message in prompt:
            role = message["role"]
            content = message["content"]

            # Map OpenAI roles to Gemini roles
            if role == "system":
                history.append({"role": "user", "parts": [content]})
            elif role == "assistant":
                history.append({"role": "model", "parts": [content]})
            elif role == "user":
                history.append({"role": "user", "parts": [content]})

        chat = self.__model.start_chat(history=history[:-1])

        response = chat.send_message(
            history[-1]["parts"][0],
            tools=tool_declarations,
        )

        return response



from __future__ import annotations

import os
from threading import Lock, Thread
from typing import Dict, List, Optional, Union

from arkaine.internal.registrar import Registrar
from arkaine.internal.store.context import Check, ContextStore, Query
from arkaine.tools.context import Context


class GlobalContextStore(ContextStore):
    __store: ContextStore = None
    __lock = Lock()
    __enabled = False
    __thread_lock = Lock()
    __thread: Thread = None

    def __init__(self):
        raise ValueError("GlobalContextStore cannot be instantiated")

    @classmethod
    def set_store(cls, store: ContextStore):
        with cls.__lock:
            cls.__store = store

    @classmethod
    def get_store(cls) -> ContextStore:
        with cls.__lock:
            return cls.__store

    @classmethod
    def enable_autosave(cls):
        Registrar.add_on_tool_call(cls._autosave)
        Registrar.add_on_llm_call(cls._autosave)

    @classmethod
    def _autosave(cls, _, ctx: Context):
        with cls.__lock:
            if not cls.__enabled:
                return
            cls.__store.save(ctx)

            # When the context is finished, save it.
            # TODO - this needs a lot of work to deal with
            # timing issues!
            ctx.add_on_end_listener(lambda ctx: cls.__store.save(ctx))

    @classmethod
    def disable_autosave(cls):
        with cls.__lock:
            if not cls.__enabled:
                return
            cls.__enabled = False


class InMemoryContextStore(ContextStore):
    """
    MemoryContextStore is a simple in-memory context store.
    """

    def __init__(self, contexts: Dict[str, Context] = {}):
        super().__init__()
        self._contexts: Dict[str, Context] = contexts

    def get_context(self, id: str) -> Optional[Context]:
        return self._contexts.get(id)

    def query_contexts(
        self, query: Union[Query, List[Query], Check, List[Check]]
    ) -> List[Context]:
        if isinstance(query, List):
            q = Query()
            for i in query:
                q += i
            query = q
        elif isinstance(query, Check):
            query = Query([query])

        return [
            context for context in self._contexts.values() if query(context)
        ]

    def save(self, context: Context) -> None:
        self._contexts[context.id] = context

    @staticmethod
    def load(self) -> "ContextStore":
        raise NotImplementedError("InMemoryContextStore is not persistent")

    def __enter__(self) -> InMemoryContextStore:
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        pass


class FileContextStore(InMemoryContextStore):

    def __init__(self, folder_path: str, contexts: Dict[str, Context]):
        super().__init__()
        self._folder_path = folder_path
        self._contexts = contexts

    @staticmethod
    def load(cls, folder_path: str) -> "ContextStore":
        contexts = {}
        for root, _, files in os.walk(folder_path):
            for file in files:
                file_path = os.path.join(root, file)
                context = Context.load(file_path)
                contexts[context.id] = context

        return cls(folder_path, contexts)

    def save(self, context: Context) -> None:
        super().save()
        context.save(os.path.join(self._folder_path, context.id))



import re
from typing import List


def isolate_sentences(text: str) -> List[str]:
    sentences = []
    current_sentence = ""
    for word in text.split():
        current_sentence += word.strip() + " "
        if word.endswith(".") or word.endswith("?") or word.endswith("!"):
            sentences.append(current_sentence.strip())
            current_sentence = ""

    return sentences


def chunk_text_by_sentences(
    text: str,
    sentences_per: int,
    overlap: int = 0,
    isolate_paragraphs: bool = False,
):
    if isolate_paragraphs:
        paragraphs = re.split(r"\n{2,}", text)
        text = [p.strip() for p in paragraphs if p.strip()]
    else:
        text = [text]

    chunks = []

    for paragraph in text:
        sentences = isolate_sentences(paragraph)
        while len(sentences) > 0:
            chunks.append(" ".join(sentences[0:sentences_per]))
            sentences = sentences[sentences_per + 1 - overlap :]

    return chunks



from typing import Callable, Optional, Union
from uuid import uuid4


class Resource:
    def __init__(
        self,
        source: str,
        name: str,
        type: str,
        description: str,
        content: Union[str, Callable[[], str]],
        id: Optional[str] = None,
    ):
        self.id = id if id else str(uuid4())
        self.name = name
        self.source = source
        self.type = type
        self.description = description
        self.__content = content

    @property
    def content(self):
        if callable(self.__content):
            return self.__content()
        return self.__content

    def to_json(self):
        return {
            "id": self.id,
            "name": self.name,
            "source": self.source,
            "type": self.type,
            "description": self.description,
        }

    @classmethod
    def from_json(cls, json: dict):
        return cls(
            json["source"],
            json["name"],
            json["type"],
            json["description"],
            json["id"],
        )

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, Resource):
            return False
        return self.id == value.id

    def __str__(self):
        return (
            f"ID: {self.id}\n"
            f"NAME: {self.name}\n"
            f"TYPE: {self.type}\n"
            f"SOURCE: {self.source}\n"
            f"DESCRIPTION: {self.description}"
        )

    def __repr__(self):
        return self.__str__()

    def __getstate__(self):
        # Ensure content is evaluated before pickling
        state = self.__dict__.copy()
        if callable(state["_Resource__content"]):
            try:
                state["_Resource__content"] = state["_Resource__content"]()
            except Exception as e:
                print(f"Error evaluating content: {e}")
                state["_Resource__content"] = ""
        return state



from arkaine.flow.branch import Branch
from arkaine.flow.conditional import Conditional
from arkaine.flow.fire_and_forget import FireAndForget
from arkaine.flow.linear import Linear
from arkaine.flow.on_error import OnError
from arkaine.flow.parallel_list import ParallelList
from arkaine.flow.retry import Retry
from arkaine.flow.dowhile import DoWhile

__all__ = [
    "Linear",
    "Branch",
    "Conditional",
    "ParallelList",
    "Retry",
    "OnError",
    "FireAndForget",
    "DoWhile",
]



import json
import re
from dataclasses import dataclass, field
from typing import Dict, List, Union


@dataclass
class Label:
    name: str = field(metadata={"transform": str.lower})
    required: bool = False
    data_type: str = "text"
    requires: List[str] = field(default_factory=list)
    required_with: List[str] = field(default_factory=list)
    is_json: bool = False


class Parser:
    """
    A flexible text parser that extracts labeled sections from text input.

    The parser identifies sections by looking for labels followed by separators
    (colon, tilde, or dash) and captures the content that follows. Labels are
    case-insensitive and can contain multiple words.

    Basic usage:
    ```python
    # Create parser with simple string labels
    parser = Parser(['name', 'description', 'requirements'])

    # Or use Label objects for more control
    parser = Parser([
        Label(name='name', required=True),
        Label(name='description', data_type='text'),
        Label(name='config', is_json=True)
    ])

    # Parse text
    result = parser.parse('''
        Name: John Smith
        Description: A software engineer
        Config: {"level": "senior"}
    ''')
    ```

    Features:
    - Case-insensitive label matching
    - Multiple separator support (`:`, `~`, `-`)
    - Multi-line content capture
    - Required field validation
    - JSON field parsing
    - Dependency validation between fields
    - Flexible label definitions with metadata

    Args:
        labels (List[Union[str, Label]]): List of labels to parse for. Can be
            simple strings or Label objects for more control. Label objects
            support additional features like required fields, JSON parsing,
            and field dependencies.

    Returns:
        Dict containing:
            - data: Dictionary of parsed values keyed by label name
            - errors: List of validation errors if any occurred
    """

    def __init__(self, labels: List[Union[str, Label]]):
        for index, label in enumerate(labels):
            if isinstance(label, str):
                labels[index] = Label(name=label.lower())

        # Sort labels by descending length to prevent partial matches
        self.__labels = sorted(labels, key=lambda x: -len(x.name))
        # Store lowercase keys in label map
        self.__label_map = {
            label.name.lower(): label for label in self.__labels
        }
        self.__patterns = self._build_patterns()

    def _build_patterns(self):
        patterns = []
        for label in self.__labels:
            # Replace spaces in label names with \s+ to allow multiple spaces
            label_regex = r"\s+".join(map(re.escape, label.name.split(" ")))
            # Require at least one colon/tilde/dash before treating it as a label
            pattern = re.compile(
                r"^\s*" + label_regex + r"\s*[:~\-]+\s*",
                re.IGNORECASE,
            )
            patterns.append((label.name, pattern))
        return patterns

    def parse(self, text: str) -> Dict:
        """
        Parses the text into a structure of lists (one list per label).
        This original parse method lumps together values for each label into arrays.
        """
        text = self._clean_text(text)
        lines = [line.rstrip() for line in text.split("\n")]

        raw_data = {label.name: [] for label in self.__labels}
        current_label = None
        current_entry = ""

        for line in lines:
            label_name, value = self._parse_line(line)

            if label_name:
                if current_label:
                    self._finalize_entry(raw_data, current_label, current_entry)
                current_label = label_name
                current_entry = value
            else:
                if current_label:
                    # Append the line to the current entry, ensuring we handle new lines correctly
                    current_entry += (
                        "\n" + line.strip() if current_entry else line.strip()
                    )

        if current_label:
            self._finalize_entry(raw_data, current_label, current_entry)

        return self._process_results(raw_data)

    def parse_blocks(self, text: str, block_label: str) -> List[Dict]:
        """
        Parses the text into multiple blocks. Each block starts whenever the specified
        'block_label' is encountered. The label definitions and validations still apply,
        but the result is now a list of dictionaries (one per block).

        Example:
            blocks = parser.parse_blocks(text, 'resource')
            # => [
            #     {"resource": "...", "reason": "...", "recommended": "..."},
            #     {"resource": "...", "reason": "...", "recommended": "..."},
            #     ...
            # ]

        Args:
            text (str): The text to parse.
            block_label (str): The label that signifies the start of a new block.

        Returns:
            List[Dict]: A list of parsed blocks, where each block is
                        a dict of label -> list-of-values or single value
        """
        text = self._clean_text(text)
        lines = [line.rstrip() for line in text.split("\n")]
        block_label_lower = block_label.lower()

        blocks = []
        raw_data = {label.name: [] for label in self.__labels}
        current_label = None
        current_entry = ""
        have_any_data = False

        for line in lines:
            label_name, value = self._parse_line(line)

            if label_name:
                # (1) First finalize the previous label for the "old" block
                if current_label is not None:
                    self._finalize_entry(raw_data, current_label, current_entry)

                # (2) If it's a new block label and we already have data, finalize the old block
                if label_name.lower() == block_label_lower and have_any_data:
                    # Debug: see raw_data for the old block
                    blocks.append(self._process_results(raw_data))
                    raw_data = {label.name: [] for label in self.__labels}

                # (3) Now set up the new label
                current_label = label_name
                current_entry = value
                have_any_data = True
            else:
                # no label => treat as continuation
                if current_label:
                    current_entry += (
                        "\n" + line.strip() if current_entry else line.strip()
                    )

        # end of loop: finalize the last label in the last block
        if current_label:
            self._finalize_entry(raw_data, current_label, current_entry)
        if have_any_data:
            blocks.append(self._process_results(raw_data))

        return blocks

    def _clean_text(self, text):
        # First, handle any code blocks
        def extract_content(match):
            return match.group(1)  # Just return the content inside

        # Handle all code blocks, regardless of language
        text = re.sub(
            r"```(?:\w+)?\s*(.*?)\s*```", extract_content, text, flags=re.DOTALL
        )

        # Remove inline code markers
        text = re.sub(r"`([^`]+)`", r"\1", text)

        # # Remove leading non-word characters (but preserve colons)
        # text = re.sub(r"^\s*([^:\w]*)\b", "", text, flags=re.MULTILINE)

        return text.strip()

    def _parse_line(self, line):
        # 1) Try pattern list
        for label_name, pattern in self.__patterns:
            if match := pattern.match(line):
                value_start = match.end()
                value = line[value_start:].strip()
                return label_name, value

        # 2) Fallback
        for label_name in self.__label_map:
            if line.strip().lower().startswith(label_name.lower()):
                remain = line.strip()[len(label_name) :]
                if re.match(r"^\s*[:~\-]+", remain):
                    # returns label
                    content = re.sub(r"^\s*[:~\-]+", "", remain).strip()
                    return label_name, content
                else:
                    # treat as continuation
                    return None, line.strip()

        # No match; treat as continuation
        return None, None

    def _finalize_entry(self, data, label_name, entry):
        if entry is None:
            entry = ""
        content = entry.strip()

        if content:
            if label_name not in data:
                data[label_name] = []
            data[label_name].append(content)

    def _process_results(self, raw_data):
        processed = {}
        errors = []

        for label_name in raw_data:
            label_def = self.__label_map[label_name.lower()]
            # Ensure label name is lowercase in output
            processed[label_name.lower()] = []

            for entry in raw_data[label_name]:
                processed_entry, error = self._process_entry(label_def, entry)
                processed[label_name.lower()].append(processed_entry)
                if error:
                    errors.append(error)

        errors += self._validate_dependencies(processed)
        return {"data": processed, "errors": errors}

    def _process_entry(self, label_def, entry):
        if label_def.is_json:
            try:
                return json.loads(entry), None
            except json.JSONDecodeError as e:
                return entry, f"JSON error in '{label_def.name}': {str(e)}"
        return entry, None

    def _validate_dependencies(self, data):
        errors = []
        for label_name, entries in data.items():
            # Use lowercase key to look up label definition
            label_def = self.__label_map[label_name.lower()]

            if label_def.required and not entries:
                errors.append(f"Required label '{label_name}' missing")

            if entries:
                for req in label_def.required_with:
                    # Use lowercase when checking required dependencies
                    if not data.get(req.lower(), []):
                        errors.append(f"'{label_name}' requires '{req}'")
        return errors



from typing import Any, Dict


class Finding:

    def __init__(self, source: str, summary: str, content: str):
        self.source = source
        self.summary = summary
        self.content = content

    def to_json(self):
        return {
            "source": self.source,
            "content": self.content,
            "summary": self.summary,
        }

    @classmethod
    def from_json(cls, json: Dict[str, Any]):
        return cls(json["source"], json["summary"], json["content"])

    def __str__(self):
        return f"{self.source}\n{self.summary}\n{self.content}"

    def __repr__(self):
        return self.__str__()



"""Logger module for tracking and displaying context execution and events."""

import json
import sys
from threading import Lock
from typing import Any, Dict, Optional, TextIO

from arkaine.internal.registrar import Registrar
from arkaine.tools.context import Attachable, Context
from arkaine.tools.events import Event
from arkaine.tools.tool import Tool


class Colors:
    """ANSI color codes for terminal output formatting."""

    RESET = "\033[0m"
    BOLD = "\033[1m"
    RED = "\033[31m"
    GREEN = "\033[32m"
    YELLOW = "\033[33m"
    BLUE = "\033[34m"
    MAGENTA = "\033[35m"
    CYAN = "\033[36m"
    GRAY = "\033[90m"


class Logger:
    """Thread-safe logger for context execution and events.

    Provides real-time and post-execution logging of contexts and their events,
    with support for colored output and JSON export.
    """

    def __init__(
        self,
        output_stream: Optional[TextIO] = sys.stdout,
        use_colors: bool = True,
        indent_size: int = 2,
        event_colors: Optional[Dict[str, str]] = None,
    ):
        """Initialize the logger with specified configuration.

        Args:
            output_stream: Where to write log output. Defaults to stdout.
            use_colors: Whether to use ANSI colors in output. Defaults to True.
            indent_size: Number of spaces for JSON indentation. Defaults to 2.
            event_colors: Custom color mapping for event types.
        """
        self.output_stream = output_stream
        self.use_colors = use_colors
        self.indent_size = indent_size
        self._lock = Lock()
        self._event_colors = (
            event_colors
            if event_colors
            else {
                "child_context_created": Colors.GREEN,
                "context_update": Colors.BLUE,
                "context_exception": Colors.RED,
                "context_output": Colors.CYAN,
                "agent_called": Colors.YELLOW,
                "agent_prompt": Colors.MAGENTA,
                "agent_llm_response": Colors.CYAN,
                "tool_called": Colors.YELLOW,
                "tool_return": Colors.GREEN,
            }
        )

    def attach_tool(self, tool: Tool):
        """
        Attach a tool to the logger such that whenever the tool is utilized the
        logger will appropriately log the resulting context and all sub events.
        """
        tool.add_on_call_listener(self.on_tool_call)

    def on_producer_call(self, producer: Attachable, context: Context):
        # Subscribe to the context events
        context.add_event_listener(self.log_event, ignore_children_events=True)

    def _colorize(self, text: str, color: str) -> str:
        """Apply ANSI color codes to text if colors are enabled.

        Args:
            text: The text to colorize. color: The ANSI color code to apply.

        Returns:
            The text with color codes if colors are enabled, otherwise
            unchanged.
        """
        if self.use_colors:
            return f"{color}{text}{Colors.RESET}"
        return text

    def _format_data(self, data: Any) -> str:
        """
        Format data for display, using JSON formatting when possible.

        Args:
            data: The data to format.

        Returns:
            A string representation of the data.
        """
        if data is None:
            return ""

        if isinstance(data, (dict, list)):
            try:
                return json.dumps(data, indent=self.indent_size)
            except (TypeError, ValueError):
                return str(data)
        return str(data)

    def _format_event(self, event: Event, context: Context) -> str:
        """
        Format an event for display with context information.

        Args:
            event: The event to format. context: The context object containing
            the event.

        Returns:
            A formatted string representing the event.
        """
        color = self._event_colors.get(event._event_type, Colors.GRAY)
        timestamp = event._get_readable_timestamp()
        event_type = event._event_type.replace("_", " ").title()

        # Build the header with context information
        header = f"{timestamp} - {event_type}"

        # Add tool name for tool-related events
        header += f" [{context.attached.name} | {context.attached.id}]"

        # Format the event data
        if event.data:
            # Use to_json if available
            if hasattr(event.data, "to_json"):
                formatted_data = self._format_data(event.data.to_json())
            else:
                formatted_data = self._format_data(event.data)

            if formatted_data:
                if "\n" in formatted_data:
                    indented_data = "\n".join(
                        f"    {line}" for line in formatted_data.split("\n")
                    )
                    result = f"{header}:\n{indented_data}"
                else:
                    result = f"{header}:\n    {formatted_data}"
            else:
                result = header
        else:
            result = header

        return self._colorize(result, color)

    def _write(self, text: str):
        """Write text to output stream in a thread-safe manner.

        Args:
            text: The text to write.
        """
        if self.output_stream:
            with self._lock:
                self.output_stream.write(f"{text}\n")
                self.output_stream.flush()

    def log_event(self, context: Context, event: Event):
        """Log a single event with optional context information.

        Args:
            event: The event to log.
            context: The executing context
        """
        formatted = self._format_event(event, context)
        self._write(formatted)

    def cleanup(self):
        self.output_stream = None

    def __del__(self):
        self.cleanup()


class GlobalLogger:
    """
    Global logger is just a singleton instance of a logger
    that utilizes the registrar to automatically detect and
    thus log all tools.
    """

    _instance: Logger = None

    def __init__(self):
        raise ValueError("GlobalLogger is a singleton")

    @classmethod
    def get_instance(cls):
        if not cls._instance:
            cls._instance = Logger()
        return cls._instance

    @classmethod
    def enable(cls):
        Registrar.enable()
        instance = cls.get_instance()
        Registrar.add_on_producer_call(instance.on_producer_call)

    @classmethod
    def disable(cls):
        del cls._instance



from __future__ import annotations

import json
import os
from typing import Any, Dict, List, Optional, Tuple

import openai as oaiapi
from openai.types.chat.chat_completion import ChatCompletion

from arkaine.backends.backend import Backend
from arkaine.backends.common import simple_tool_results_to_prompts
from arkaine.tools.agent import Prompt
from arkaine.tools.context import Context
from arkaine.tools.tool import Tool
from arkaine.tools.types import ToolArguments, ToolResults
from arkaine.utils.templater import PromptTemplate
from arkaine.utils.tool_format import openai as openai_tool_format


class OpenAI(Backend):

    def __init__(
        self,
        tools: List[Tool] = [],
        template: PromptTemplate = PromptTemplate.default(),
        max_simultaneous_tools: int = -1,
        api_key: Optional[str] = None,
        model: str = "gpt-3.5-turbo",
        temperature: float = 0.7,
        max_tokens: int = 1024,
        initial_state: Dict[str, Any] = {},
    ):
        super().__init__(None, tools, max_simultaneous_tools, initial_state)
        self.template = template
        self.model = model
        self.temperature = temperature
        self.max_tokens = max_tokens
        api_key = api_key or os.getenv("OPENAI_API_KEY")
        self.__client = oaiapi.Client(api_key=api_key)

    def parse_for_result(
        self, context: Context, response: ChatCompletion
    ) -> str:
        return response.choices[0].message.content

    def parse_for_tool_calls(
        self,
        context: Context,
        response: ChatCompletion,
        stop_at_first_tool: bool = False,
    ) -> List[Tuple[str, ToolArguments]]:
        """
        parse_for_tool_calls accepts a chatgpt response, extract all tool
        calls.

        Note that there is no "stop_at_first_tool" functionality on this
        function like other backends.
        """
        tool_calls: List[Dict[str, Any]] = []
        if response.choices[0].message.tool_calls:
            for tool_msg in response.choices[0].message.tool_calls:
                tool_name = tool_msg.function.name
                params = json.loads(tool_msg.function.arguments)

                tool_calls.append((tool_name, params))

        return tool_calls

    def tool_results_to_prompts(
        self,
        context: Context,
        prompt: Prompt,
        results: ToolResults,
    ) -> List[Prompt]:
        return simple_tool_results_to_prompts(prompt, results)

    def prepare_prompt(self, context: Context, **kwargs) -> Prompt:
        return self.template.render(kwargs)

    def query_model(self, context: Context, prompt: Prompt) -> ChatCompletion:
        return self.__client.chat.completions.create(
            model=self.model,
            messages=prompt,
            max_tokens=self.max_tokens,
            temperature=self.temperature,
            tools=[openai_tool_format(tool) for tool in self.tools.values()],
        )



from arkaine.connectors.api import API
from arkaine.connectors.cli import CLI
from arkaine.connectors.rss import RSS, Feed
from arkaine.connectors.schedule import Schedule, Task
from arkaine.utils.interval import Interval

__all__ = [
    "API",
    "CLI",
    "Schedule",
    "Task",
    "Interval",
    "RSS",
    "Feed",
    "Inbox",
]



import os
from typing import Optional

from groq import Groq as GroqAPI

from arkaine.llms.llm import LLM
from arkaine.tools.agent import Prompt


class Groq(LLM):

    MODELS = {
        "gemma2-9b-it": {"context_length": 8192},
        "llama-3.3-70b-versatile": {"context_length": 128000},
        "llama-3.1-8b-instant": {"context_length": 8192},
        "llama-guard-3-8b": {"context_length": 8192},
        "llama3-70b-8192": {"context_length": 8192},
        "llama3-8b-8192": {"context_length": 8192},
        "mixtral-8x7b-32768": {"context_length": 32768},
        "llama-3.2-1b-preview": {"context_length": 128000},
        "llama-3.2-3b-preview": {"context_length": 128000},
        "llama-3.2-11b-vision-preview": {"context_length": 128000},
        "llama-3.2-90b-vision-preview": {"context_length": 128000},
        "deepseek-r1-distill-llama-70b": {"context_length": 128000},
    }

    def __init__(
        self,
        model: str = "llama3-70b-8192",
        api_key: Optional[str] = None,
        context_length: Optional[int] = 8192,
    ):
        if api_key is None:
            api_key = os.environ.get("GROQ_API_KEY")

        self.__client = GroqAPI(api_key=api_key)
        self.__model = model

        if context_length:
            self.__context_length = context_length
        elif model in self.MODELS:
            self.__context_length = self.MODELS[model]["context_length"]
        else:
            raise ValueError(
                f"Unknown model: {model} - must specify context length"
            )

        super().__init__(name=f"groq:{model}")

    @property
    def context_length(self) -> int:
        return self.__context_length

    def completion(self, prompt: Prompt) -> str:
        if isinstance(prompt, str):
            prompt = [
                {
                    "role": "system",
                    "content": prompt,
                }
            ]

        response = self.__client.chat.completions.create(
            model=self.__model,
            messages=prompt,
        )

        return response.choices[0].message.content

    def __str__(self) -> str:
        return self.name



from datetime import datetime


class Timer:
    def __init__(self):
        self.__elapsed: float = 0
        self._start_time = None

    def __enter__(self):
        self._start_time = datetime.now()
        return self

    def __exit__(self, *args):
        end_time = datetime.now()
        self.__elapsed = (end_time - self._start_time).total_seconds()

    @property
    def elapsed(self) -> float:
        return self.__elapsed



from __future__ import annotations

import inspect
import random
import re
import string
from functools import wraps
from typing import Any, Callable, List, Optional, _GenericAlias, get_type_hints

from arkaine.tools.tool import Argument, Tool


def toolify(
    tool_name: Optional[str] = None, tool_description: Optional[str] = None
):
    """
    Decorator that converts a function into a Tool object.
    """

    def decorator(func: Callable) -> Tool:
        # Get function signature
        sig = inspect.signature(func)

        # Get type hints
        type_hints = get_type_hints(func)

        # Parse docstring
        doc_description, arg_descriptions, return_desc, doc_name = (
            _parse_docstring(func.__doc__)
        )

        # Determine name priority: explicit tool_name > docstring name > function name > random lambda name
        if func.__name__ == "<lambda>":
            name = tool_name or doc_name or f"lambda_{_generate_random_id()}"
        else:
            name = tool_name or doc_name or func.__name__

        # Use provided name/description or defaults
        description = tool_description or doc_description

        # Add return description to tool description if available
        if return_desc:
            description = (
                f"{description}\n\nReturns: {return_desc}"
                if description
                else f"Returns: {return_desc}"
            )
        elif not description:
            description = f"Tool for {func.__name__}"

        # Create Arguments list
        arguments: List[Argument] = []

        for param_name, param in sig.parameters.items():
            # Skip self/cls for methods
            if param_name in ("self", "cls", "context"):
                continue

            # Determine if argument is required
            required = param.default == inspect.Parameter.empty

            # Get default value if exists
            default = None if required else str(param.default)

            # Get type as string
            param_type = type_hints.get(param_name, "Any")
            type_str = _get_full_type_str(param_type)

            # Get description from docstring if available
            param_desc = arg_descriptions.get(
                param_name, f"Parameter {param_name}"
            )

            # Create Argument object
            arg = Argument(
                name=param_name,
                description=param_desc,
                type=type_str,
                required=required,
                default=default,
            )
            arguments.append(arg)

        # Create and return Tool
        @wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        tool = Tool(
            name=name,
            description=description,
            args=arguments,
            func=wrapper,
        )

        return tool

    # Handle case where decorator is used without parentheses
    if callable(tool_name):
        func = tool_name
        tool_name = None
        return decorator(func)

    return decorator


def _parse_docstring(
    docstring: Optional[str],
) -> tuple[str, dict[str, str], Optional[str], Optional[str]]:
    """
    Parse a docstring to extract the name, description, argument descriptions,
    and return description.
    """
    if not docstring:
        return "", {}, None, None

    # Clean and split the docstring
    lines = [line.strip() for line in docstring.split("\n")]
    if not lines:
        return "", {}, None, None

    # Look for Name: tag
    name = None
    cleaned_lines = []
    for line in lines:
        if line.lower().startswith("name:"):
            name = line[5:].strip()
        else:
            cleaned_lines.append(line)

    # Process the rest of the docstring with cleaned lines
    lines = cleaned_lines

    description = []
    arg_descriptions: dict[str, str] = {}
    return_description = None

    # Find where the parameters section begins
    param_start_idx = len(lines)
    param_indicators = [
        ":param",
        "Args:",
        "Parameters:",
        "Arguments:",
    ]

    plain_param_pattern = re.compile(r"\s*\w+\s+(?:--|—|-)\s+.+")

    # First pass: find the first parameter section and description
    for idx, line in enumerate(lines):
        line = line.strip()
        if not line:
            continue
        if (
            any(line.startswith(ind) for ind in param_indicators)
            or line in param_indicators
            or plain_param_pattern.match(line)
        ):
            param_start_idx = idx
            break
        description.append(line)

    # Track the format being used to maintain consistency
    current_format = None  # "rst", "google", or "plain"
    in_section = False
    current_param = None
    current_desc = []
    in_returns = False

    # Second pass: process parameters and returns
    for line in lines[param_start_idx:]:
        line = line.strip()
        if not line:
            continue

        # Detect format if not yet determined
        if not current_format:
            if line.startswith(":param"):
                current_format = "rst"
            elif line in ["Args:", "Parameters:", "Arguments:"]:
                current_format = "google"
                in_section = True
                continue
            elif plain_param_pattern.match(line):
                current_format = "plain"

        # Handle return descriptions
        if any(
            line.startswith(ind)
            for ind in [":return", ":returns", "Returns:", "Return:"]
        ) or line.lower().startswith(("returns --", "return --")):
            if current_param and current_desc:
                arg_descriptions[current_param] = " ".join(current_desc)
            current_param = None
            current_desc = []
            in_returns = True

            return_match = re.match(
                r"(?::returns?|Returns?:|\w+\s*--)\s*(.+)?", line
            )
            if (
                return_match
                and return_match.group(1)
                and not return_description
            ):
                return_description = return_match.group(1)
            continue

        # Handle return continuation lines
        if in_returns and not return_description:
            return_description = line
            continue
        elif (
            in_returns
            and return_description
            and not any(line.startswith(ind) for ind in param_indicators)
        ):
            return_description += f" {line}"
            continue

        # Parse according to format
        if current_format == "rst":
            rst_match = re.match(r":param\s+(\w+)\s*:\s*(.+)", line)
            if rst_match:
                if current_param:
                    arg_descriptions[current_param] = " ".join(current_desc)
                current_param = rst_match.group(1)
                current_desc = [rst_match.group(2)]
                in_returns = False
            elif current_param and not line.startswith(":"):
                current_desc.append(line)

        elif current_format == "google" and in_section:
            if line in ["Returns:", "Return:"]:
                in_section = False
                continue
            google_match = re.match(
                r"\s*(\w+)(?:\s*\([^)]+\))?\s*:\s*(.+)", line
            )
            if google_match:
                if current_param:
                    arg_descriptions[current_param] = " ".join(current_desc)
                current_param = google_match.group(1)
                current_desc = [google_match.group(2)]
                in_returns = False
            elif current_param and not any(
                line.startswith(ind) for ind in param_indicators + ["Returns:"]
            ):
                current_desc.append(line)

        elif current_format == "plain":
            plain_match = re.match(r"\s*(\w+)\s*(?:--|—|-)\s*(.+)", line)
            if plain_match:
                if current_param:
                    arg_descriptions[current_param] = " ".join(current_desc)
                current_param = plain_match.group(1)
                current_desc = [plain_match.group(2)]
                in_returns = False

    # Don't forget to add the last parameter
    if current_param and current_desc:
        arg_descriptions[current_param] = " ".join(current_desc)

    # Clean up descriptions
    description = " ".join(description).strip()
    arg_descriptions = {k: v.strip() for k, v in arg_descriptions.items()}
    if return_description:
        return_description = return_description.strip()

    return description, arg_descriptions, return_description, name


def _get_full_type_str(type_hint: Any) -> str:
    """
    Get a string representation of a type hint, including generic parameters.
    """
    if isinstance(type_hint, _GenericAlias):
        return str(type_hint).replace("typing.", "")
    return getattr(type_hint, "__name__", str(type_hint))


def _generate_random_id(n: int = 6) -> str:
    """Generate a random n-character alphanumeric string (lowercase)."""
    return "".join(random.choices(string.ascii_lowercase + string.digits, k=n))



import time
from typing import Any, List, Optional, Tuple, Type, Union

from arkaine.tools.context import Context
from arkaine.tools.tool import Tool


class Retry(Tool):
    """
    A wrapper tool that retries failed executions of another tool.

    This tool will catch exceptions and retry the execution up to a specified
    number of times. It can be configured to only retry on specific exception
    types and to wait between retries.

    Args:
        tool (Tool): The base tool to wrap and retry on failure
        max_retries (int): Maximum number of retry attempts (not counting
        initial try)
        exceptions (Union[Type[Exception], List[Type[Exception]]], optional):
            Specific exception type(s) to retry on. If None, retries on any
            Exception.
        delay (float, optional): Time in seconds to wait between retries.
            Defaults to 0 (no delay).
        name (str, optional): Custom name for the tool.
            Defaults to "{tool.name}::retry_{max_retries}".
        description (str, optional): Custom description.
            Defaults to base tool's description.
    """

    def __init__(
        self,
        tool: Tool,
        max_retries: int,
        exceptions: Optional[
            Union[Type[Exception], List[Type[Exception]]]
        ] = None,
        delay: float = 0,
        name: Optional[str] = None,
        description: Optional[str] = None,
    ):
        self._tool = tool
        self._max_retries = max_retries
        self._delay = delay

        # Handle single exception or list of exceptions
        if exceptions is None:
            self._exceptions = (Exception,)
        elif isinstance(exceptions, list):
            self._exceptions = tuple(exceptions)
        else:
            self._exceptions = (exceptions,)

        if not name:
            name = f"{tool.name}::retry_{max_retries}"

        if not description:
            description = tool.description

        super().__init__(
            name=name,
            args=tool.args,
            description=description,
            func=self.retry_invoke,
            examples=tool.examples,
        )

    def retry_invoke(self, context: Context, **kwargs):
        """
        Execute the wrapped tool with retry logic.

        Args:
            context (Context): The execution context

            **kwargs: Arguments to pass to the wrapped tool

        Returns:
            Any: The result from a successful execution

        Raises:
            Exception: The last exception encountered after all retries are
            exhausted
        """
        attempts = 0
        last_exception = None

        while attempts <= self._max_retries:
            context["attempt"] = attempts + 1
            try:
                return self._tool(context, **kwargs)
            except self._exceptions as e:
                last_exception = e
                attempts += 1

                if attempts <= self._max_retries:
                    if self._delay > 0:
                        time.sleep(self._delay)
                    continue
                break

        # If we get here, we've exhausted all retries
        raise last_exception

    def retry(self, context: Context) -> Any:
        if context.attached is None:
            raise ValueError("no tool assigned to context")

        if context.attached != self:
            raise ValueError(
                f"context is not for {self.name}, is instead for "
                f"{context.attached.name}"
            )

        if context.children is None:
            # In this branch, we never actually succeeded in running the tool
            # and thus never generated any children; therefore we stop here
            # and just re-call.
            original_output = context.output
            context.clear(executing=True)

            return self(context, original_output)
        else:
            # In this branch, we have already generated children and thus
            # we can just retry the last child tool.
            child_ctx = context.children[-1]
            context.children = []

            return self._tool.retry(child_ctx)


def retry(
    max_retries: int = 3,
    delay: float = 0,
    exceptions: Union[Type[Exception], Tuple[Type[Exception], ...]] = Exception,
    name: Optional[str] = None,
    description: Optional[str] = None,
):
    """
    A decorator that wraps a Tool to add retry functionality.

    Args:
        max_retries (int): Maximum number of retry attempts. Defaults to 3.
        delay (float): Delay in seconds between retries. Defaults to 0.
        exceptions (Union[Type[Exception], Tuple[Type[Exception], ...]]):
            Exception type(s) to catch and retry on. Defaults to Exception.
    """

    def decorator(tool: Tool):
        # Save original invoke method

        return Retry(
            tool,
            max_retries,
            exceptions,
            delay,
            name,
            description,
        )

    return decorator



from typing import Any, Dict, List, Optional, Tuple

from arkaine.internal.store.embeddings import InMemoryEmbeddingStore
from arkaine.tools.tool import Argument, Context, Tool
from arkaine.tools.wrapper import Wrapper
from arkaine.utils.documents import cosine_distance
from arkaine.utils.embeddings.model import OllamaEmbeddingModel


class ContentFilter(Tool):
    """
    A tool that filters a body of text based on semantic similarity to a query.
    Can optionally cluster results and return only the highest scoring cluster.

    Args:
        name (str): Name for this filter instance. Defaults to "ContentFilter"

        description (str): Description for this filter instance.

        n (int): Number of top results to return. Defaults to 5.

        embedder (Optional[InMemoryEmbeddingStore]): Custom embedding store to
            use. If None, a fresh InMemoryEmbeddingStore is created for
            each call. If a class is provided, it will be instantiated with
            embedder_arguments for each call. If an object instance is
            provided, it will be reused for all calls (embedder_arguments
            ignored).

        embedder_arguments (Optional[Dict[str, Any]]): Arguments to initialize
            the embedder class. Required when embedder is a class. Ignored when
            embedder is None or an instance. Example: {"model":
            "text-embedding-3-small"}

        cluster_threshold (Optional[float]): If set, will cluster results and
        return
            only the highest scoring cluster. Values between 0-1, with higher
            values creating tighter clusters. Recommended range: 0.7-0.9
    """

    def __init__(
        self,
        name: str = "ContentFilter",
        description: Optional[str] = None,
        n: int = 5,
        embedding_store: Optional[InMemoryEmbeddingStore] = None,
        embedding_store_args: Optional[Dict[str, Any]] = None,
        cluster_threshold: Optional[float] = None,
    ):
        self.n = n
        self.cluster_threshold = cluster_threshold

        # Validate embedder and embedder_arguments combination
        if embedding_store is None:
            self.embedding_store = InMemoryEmbeddingStore
            self.embedding_store_args = {
                "embedding_model": OllamaEmbeddingModel()
            }
        elif isinstance(embedding_store, type):
            if embedding_store_args is None:
                raise ValueError(
                    "embedder_arguments required when embedder is a class"
                )
            self.embedding_store = embedding_store
            self.embedding_store_args = embedding_store_args
        else:
            if embedding_store_args:
                raise ValueError(
                    "embedder_arguments should be None when embedder is an"
                    + " instance"
                )
            self.embedding_store = embedding_store
            self.embedding_store_args = None

        args = [
            Argument(
                "content",
                "The text content to filter",
                "string",
                required=True,
            ),
            Argument(
                "query",
                "Query to filter results against",
                "string",
                required=True,
            ),
        ]

        if description is None:
            description = (
                f"Filters text content to the {n} most relevant results "
                "based on semantic similarity to a query."
            )
            if cluster_threshold:
                description += (
                    f" Results are clustered with threshold {cluster_threshold} "
                    "and only the highest scoring cluster is returned."
                )

        super().__init__(name, description, args, self.filter_content)

    def _process_content(self, content: Any) -> List[str]:
        """Convert various input types to list of strings for embedding."""
        if isinstance(content, str):
            # Split on newlines and filter out empty strings and whitespace-only strings
            lines = [line.strip() for line in content.split("\n")]
            return [line for line in lines if line]
        elif isinstance(content, dict):
            return [str(v) for v in content.values()]
        elif isinstance(content, (list, tuple)):
            return [str(x) for x in content]
        else:
            return [str(content)]

    def _cluster_results(
        self, results: List[Any], scores: List[float]
    ) -> List[Any]:
        """
        Cluster results based on similarity scores and return highest cluster.
        Uses a simple threshold-based clustering approach.
        """
        if not results:
            return []

        # Start with highest scoring result
        clusters = [[0]]  # List of clusters, each containing result indices

        # Group remaining results into clusters
        for i in range(1, len(results)):
            score_diff = abs(scores[0] - scores[i])

            if score_diff <= self.cluster_threshold:
                clusters[0].append(i)
            else:
                # Start new cluster
                clusters.append([i])

        # Return items from highest scoring cluster
        return [results[i] for i in clusters[0]]

    def filter_content(self, content: str, query: str) -> List[str]:
        """Filter content based on similarity to query."""
        items = self._process_content(content)

        if self.embedding_store_args is not None:
            embedder = self.embedding_store(**self.embedding_store_args)
        else:
            embedder = self.embedding_store

        embedder.add_text(items)

        # Get top N results
        filtered = embedder.query(query, top_n=self.n)

        if self.cluster_threshold:
            # Get similarity scores for clustering
            query_embedding = embedder.get_embedding(query)
            scores = [
                cosine_distance(query_embedding, embedder.get_embedding(item))
                for item in filtered
            ]
            filtered = self._cluster_results(filtered, scores)

        return filtered


class ContentFilterWrapper(Wrapper):
    """
    A wrapper that filters another tool's output using semantic similarity.

    Args:
        tool (Tool): The tool to wrap
        name (Optional[str]): Name for the wrapper. Defaults to
            "{tool.name}::cf"
        n (int): Number of top results to return
        embedder (Optional[InMemoryEmbeddingStore]): Custom embedding store to
            use
        embedder_arguments (Optional[Dict[str, Any]]): Arguments to initialize
            the embedder class. Required when embedder is a class. Ignored when
            embedder is None or an instance. Example: {"model":
            "text-embedding-3-small"}
        cluster_threshold (Optional[float]): Clustering threshold (0-1)
        argument_name (str): Name of the argument that contains the query.
            Defaults to "query". This is to be used when the wrapped tool
            takes multiple arguments and the query is passed in as one of them.
    """

    def __init__(
        self,
        tool: Tool,
        name: Optional[str] = None,
        n: int = 5,
        embedder: Optional[InMemoryEmbeddingStore] = None,
        embedder_arguments: Optional[Dict[str, Any]] = None,
        cluster_threshold: Optional[float] = None,
        argument_name: str = "query",
    ):
        if name is None:
            name = f"{tool.name}::cf"

        self.argument_name = argument_name

        description = (
            f"Filters the output of {tool.name} to the {n} most relevant "
            + " results based on semantic similarity to a query."
        )
        if cluster_threshold:
            description += (
                f" Results are clustered with threshold {cluster_threshold} "
                "and only the highest scoring cluster is returned."
            )

        self.content_filter = ContentFilter(
            n=n,
            embedding_store=embedder,
            embedding_store_args=embedder_arguments,
            cluster_threshold=cluster_threshold,
        )

        super().__init__(
            name,
            description,
            tool,
            [
                Argument(
                    self.argument_name,
                    "Query to filter results against",
                    "string",
                    required=True,
                )
            ],
        )

    def preprocess(self, ctx: Context, **kwargs) -> Tuple[Dict[str, Any], str]:
        """Remove query from kwargs before passing to wrapped tool."""
        filtered_kwargs = kwargs.copy()
        query = filtered_kwargs.pop(self.argument_name)
        return filtered_kwargs, query

    def postprocess(self, ctx: Context, passed: Any, result: Any) -> Any:
        """Filter the tool's output using the content filter."""
        return self.content_filter(content=result, query=passed)



import json
import os
import secrets
from abc import ABC, abstractmethod
from typing import List, Optional, Union

import jwt
import uvicorn
from fastapi import FastAPI, HTTPException, Request
from fastapi.responses import JSONResponse
from pydantic import BaseModel

from arkaine.internal.registrar import Registrar
from arkaine.tools.context import Context
from arkaine.tools.tool import Tool


class AuthRequest(BaseModel):
    tools: Union[str, List[str]]
    key: str


class AuthResponse(BaseModel):
    token: str


class Auth(ABC):
    """
    Auth is an abstract class that defines an authentication and authorization
    interface for optional use for ToolAPI.
    """

    def __init__(self):
        super().__init__()

    @abstractmethod
    def auth(self, request: Request, tool: Tool) -> bool:
        """
        Authenticate and authorize the request.
        """
        pass

    @abstractmethod
    def issue(self, request: AuthRequest) -> str:
        """
        Issue an auth token for the given tools.
        """
        pass


class API(FastAPI):
    """
    API provides a FastAPI-based HTTP interface for arkAIne tools. It allows
    tools to be exposed as REST endpoints with automatic OpenAPI documentation
    generation, authentication support, and flexible input/output handling.

    Quick Start:
        Basic usage with a single tool:
        ```python
        from arkAIne.triggers.api import ToolAPI

        # Create API with a single tool
        api = API(my_tool)
        api.serve()  # Starts server at http://localhost:8000
        ```

        Multiple tools with custom prefix:
        ```python
        # Create API with multiple tools and custom route prefix
        api = API(
            tools=[tool1, tool2, tool3],
            name="MyAPI",
            prefix="/api/v1"
        )
        api.serve(port=8080)
        ```

        With authentication:
        ```python
        from arkaine.triggers.api import API, JWTAuth

        # Create auth handler with secret and API keys
        auth = JWTAuth.from_file("auth_config.json")

        # Create authenticated API
        api = API(
            tools=my_tools,
            auth=auth
        )

        # Get auth token
        token = auth.issue(AuthRequest(tools=["tool1"], key="my-api-key"))

        # Make authenticated request
        # curl -H "Authorization: Bearer {token}" \
        #      http://localhost:8000/api/tool1
        ```

    Input Methods:
        - Query parameters: /api/tool?arg1=value1&arg2=value2
        - JSON body: POST with {"arg1": "value1", "arg2": "value2"}
        - Mixed: Both query params and JSON body (body takes precedence)

    Response Format:
        Success:
        ```json
        {
            "result": <tool output>,
            "context": <context data if requested>
        }
        ```

        Error:
        ```json
        {
            "detail": "Error message",
            "context": <context data if requested>
        }
        ```

    Headers:
        - X-Return-Context: Set to "true" to include context in response
        - X-Context-ID: Returned in response with context identifier
        - Authorization: Bearer token for authenticated endpoints
    """

    def __init__(
        self,
        tools: Union[Tool, List[Tool]],
        name: Optional[str] = None,
        description: Optional[str] = None,
        prefix: str = "/api",
        api_docs: str = "/api",
        auth: Optional[Auth] = None,
    ):
        """
        Initialize the server.

        Args:
            tools: Tool or list of tools to create endpoints for
            name: Name for the API (defaults to first tool's name)
            description: Description for the API
            prefix: Prefix for all routes (e.g., "/api")
            api_docs: URL for API documentation (set to None to disable)
        """
        self.tools = [tools] if isinstance(tools, Tool) else tools
        name = name or "Arkaine API"
        description = description or "API generated from arkAIne tools"

        super().__init__(
            title=name,
            description=description,
            docs_url=api_docs,
        )

        self._prefix = prefix
        self.auth = auth

        # Add routes for tools
        for tool in self.tools:
            self.add_tool_route(tool)

        # Add authentication middleware and route if auth is provided
        if auth:
            # Add auth endpoint
            route = "/auth"
            if self._prefix:
                route = f"{self._prefix.rstrip('/')}{route}"
            self._auth_route = route
            self.add_api_route(
                route,
                self.auth_handler,
                methods=["POST"],
                response_model=AuthResponse,
            )

            # Add auth middleware
            @self.middleware("http")
            async def auth_middleware(request: Request, call_next):
                return await self._auth_middleware(request, call_next)

    async def _auth_middleware(self, request: Request, call_next):
        """Handle authentication middleware for HTTP requests."""
        try:
            # Skip auth check for auth endpoint
            if request.url.path == self._auth_route:
                return await call_next(request)

            # Find matching tool based on path
            path = request.url.path.rstrip("/")
            if self._prefix:
                prefix = self._prefix.rstrip("/")
                if path.startswith(prefix):
                    path = path[len(prefix) :]

            path = path.lstrip("/")
            matching_tool = None
            for tool in self.tools:
                if path == tool.tname:
                    matching_tool = tool
                    break

            if matching_tool and not self.auth.auth(request, matching_tool):
                return JSONResponse(
                    status_code=401, content={"detail": "Unauthorized"}
                )

            return await call_next(request)
        except Exception as e:
            return JSONResponse(status_code=500, content={"detail": str(e)})

    def __create_endpoint_handler(self, tool: Tool):
        """Create a handler function for a tool endpoint."""

        async def handler(request: Request):
            try:
                # Get query parameters
                query_params = dict(request.query_params)

                # Get JSON body if present
                body_params = {}
                if request.headers.get("content-type") == "application/json":
                    try:
                        body_params = await request.json()
                    except json.JSONDecodeError:
                        pass

                # Combine parameters (body takes precedence)
                params = {**query_params, **body_params}

                # Convert types based on tool arguments
                converted_params = {}
                for arg in tool.args:
                    if arg.name in params:
                        try:
                            if arg.type == "int":
                                converted_params[arg.name] = int(
                                    params[arg.name]
                                )
                            elif arg.type == "float":
                                converted_params[arg.name] = float(
                                    params[arg.name]
                                )
                            elif arg.type == "bool":
                                converted_params[arg.name] = str(
                                    params[arg.name]
                                ).lower() in ["true", "1", "yes"]
                            else:
                                converted_params[arg.name] = params[arg.name]
                        except (ValueError, TypeError):
                            raise HTTPException(
                                status_code=400,
                                detail=f"Invalid type for parameter {arg.name}",
                            )
                    elif arg.required:
                        raise HTTPException(
                            status_code=400,
                            detail=f"Missing required parameter: {arg.name}",
                        )
                    elif arg.default is not None:
                        converted_params[arg.name] = arg.default

                # Execute tool
                context = Context(tool)
                try:
                    result = tool(context=context, **converted_params)
                except Exception as e:
                    if request.headers.get("X-Return-Context"):
                        raise HTTPException(
                            status_code=500,
                            detail={
                                "error": str(e),
                                "context": context.to_json(),
                            },
                        )
                    else:
                        raise HTTPException(
                            status_code=500,
                            detail={"error": str(e)},
                        )

                response = {"result": result}

                if request.headers.get("X-Return-Context"):
                    response["context"] = context.to_json()
                # Always set context ID in response header

                return JSONResponse(
                    content=response,
                    headers={
                        "X-Context-ID": context.id,
                    },
                )

            except Exception as e:
                if isinstance(e, HTTPException):
                    raise e
                raise HTTPException(status_code=500, detail=str(e))

        return handler

    def add_tool_route(
        self,
        tool: Tool,
        route: Optional[str] = None,
        method: Union[str, List[str]] = "POST",
    ):
        # Confirm that the route, if set, is valid and starts/ends
        # correctly
        if route:
            if not route.startswith("/"):
                route = f"/{route}"
            if not route.endswith("/"):
                route = f"{route}/"
        else:
            route = f"/{tool.name}/"

        # Add route prefix if specified for server
        if self._prefix:
            route = f"{self._prefix.rstrip('/')}{route}"

        # Confirm that the route isn't already used
        if route in self.routes:
            raise ValueError(f"Route {route} is already registered")

        # Confirm that all methods listed are valid methods
        if isinstance(method, str):
            method = [method]
        for m in method:
            if m not in ["GET", "POST", "PUT", "DELETE", "PATCH", "OPTIONS"]:
                raise ValueError(f"Invalid method: {m}")

        # Add route documentation
        description = f"{tool.name}\n\n{tool.description}\n\nArguments:\n"
        for arg in tool.args:
            description += f"\n- {arg.name}: {arg.description}"
            if arg.required:
                description += " (required)"
            if arg.default is not None:
                description += f" (default: {arg.default})"

        if tool.examples:
            description += "\n\nExamples:\n"
            for example in tool.examples:
                description += f"\n{example.name}:"
                if example.description:
                    description += f" {example.description}"
                description += f"\nInput: {json.dumps(example.args)}"
                if example.output:
                    description += f"\nOutput: {example.output}"

        self.add_api_route(
            route,
            self.__create_endpoint_handler(tool),
            methods=method,
            description=description,
        )

    @staticmethod
    def __strip_bearer(auth_header: Optional[str]) -> Optional[str]:
        """Helper method to strip 'Bearer ' prefix from auth header"""
        if not auth_header:
            return None
        return auth_header.replace("Bearer ", "")

    async def auth_handler(self, request: Request) -> AuthResponse:
        """Handle authentication requests and return JWT tokens"""
        try:
            # Parse the raw request body into AuthRequest model
            body = await request.json()
            auth_request = AuthRequest(**body)
            return AuthResponse(token=self.auth.issue(auth_request))
        except HTTPException:
            raise
        except Exception as e:
            raise HTTPException(status_code=500, detail=str(e))

    def serve(
        self,
        host: str = "127.0.0.1",
        port: int = 8000,
        ssl_keyfile: Optional[str] = None,
        ssl_certfile: Optional[str] = None,
        ssl_keyfile_password: Optional[str] = None,
        reload: bool = False,
        workers: Optional[int] = None,
        log_level: str = "info",
        proxy_headers: bool = True,
        forwarded_allow_ips: Optional[str] = None,
    ):
        """
        Runs the API server with configurable HTTP/WebSocket support and SSL
        options.

        Args:
            host: Bind socket to this host. Defaults to "127.0.0.1".

            port: Bind socket to this port. Defaults to 8000.

            ssl_keyfile: SSL key file path for HTTPS/WSS support.

            ssl_certfile: SSL certificate file path for HTTPS/WSS support.

            ssl_keyfile_password: Password for decrypting SSL key file.

            reload: Enable auto-reload on code changes (development only).
                Defaults to False.

            workers: Number of worker processes. Defaults to 1.

            log_level: Logging level (critical, error, warning, info, debug).
                Defaults to "info".

            proxy_headers: Enable processing of proxy headers. Defaults to
                True.

            forwarded_allow_ips: Comma separated list of IPs to trust with
                proxy headers. Defaults to the $FORWARDED_ALLOW_IPS environment
                variable or "127.0.0.1".

        Note:
            - At least one of `http` or `ws` must be True.
            - For SSL support, both ssl_keyfile and ssl_certfile must be
              provided.
            - WebSocket endpoints will be available at the same routes as HTTP
              endpoints when ws=True.
        """

        ssl_config = None
        if ssl_keyfile and ssl_certfile:
            ssl_config = {
                "keyfile": ssl_keyfile,
                "certfile": ssl_certfile,
            }
            if ssl_keyfile_password:
                ssl_config["password"] = ssl_keyfile_password

        uvicorn.run(
            self,
            host=host,
            port=port,
            ssl_keyfile=ssl_keyfile,
            ssl_certfile=ssl_certfile,
            ssl_keyfile_password=ssl_keyfile_password,
            reload=reload,
            workers=workers,
            log_level=log_level,
            proxy_headers=proxy_headers,
            forwarded_allow_ips=forwarded_allow_ips,
        )


class JWTAuth(Auth):
    """
    JSONWebTokenAuth is an implementation of Auth that uses JSON Web Tokens
    for authentication and authorization.
    """

    def __init__(self, secret: str, keys: List[str]):
        super().__init__()
        self.secret = secret
        self.keys = dict(zip(keys, [None] * len(keys)))

    def auth(self, request: Request, tool: Tool) -> bool:
        token = request.headers.get("Authorization")

        if not token:
            return False

        # Strip 'Bearer ' prefix if present
        if token.startswith("Bearer "):
            token = token[7:]

        try:
            payload = jwt.decode(token, self.secret, algorithms=["HS256"])

            if "all" in payload["tools"] or tool.tname in payload["tools"]:
                return True
            else:
                return False
        except jwt.InvalidTokenError as e:
            return False

    def issue(self, request: AuthRequest) -> str:
        """
        Issue a JWT token for the given tools.
        """
        possible_tools = Registrar.get_tools()
        tools_requested = request.tools

        if isinstance(request.tools, str):
            tools_requested = [request.tools]

        key = request.key

        # Compare the key against known keys
        if key not in self.keys:
            raise HTTPException(status_code=401, detail="Invalid key")

        # Confirm that requested tools exist
        if (isinstance(tools_requested, str) and tools_requested != "all") or (
            isinstance(tools_requested, list) and tools_requested != ["all"]
        ):
            if isinstance(tools_requested, str):
                tools_requested = [tools_requested]
            for tool_name in tools_requested:
                if tool_name not in possible_tools:
                    raise HTTPException(status_code=401, detail="Invalid tool")

        # Now that we have allowed_tools, create a JWT token
        payload = {"tools": tools_requested}
        token = jwt.encode(payload, self.secret, algorithm="HS256")

        return token

    @classmethod
    def from_env(cls):
        secret = os.getenv("JWT_SECRET")
        if not secret:
            raise ValueError("JWT_SECRET environment variable is not set")
        keys = os.getenv("JWT_KEYS")
        if not keys:
            raise ValueError("JWT_KEYS environment variable is not set")
        return cls(secret, keys)

    @classmethod
    def from_file(cls, path: str):
        with open(path, "r") as f:
            config = json.load(f)
        return cls(config["secret"], config["keys"])

    def create_key_file(self, path: str, key_count: int = 1):
        # Generate a secret and at least one key:
        secret = secrets.token_urlsafe(32)
        keys = [secrets.token_urlsafe(32) for _ in range(key_count)]
        with open(path, "w") as f:
            json.dump({"secret": secret, "keys": keys}, f)



from typing import List, Optional, Union

import googlemaps

from arkaine.tools.tool import Argument, Example, Result, Tool


class LocalSearch(Tool):
    """
    A tool for searching local businesses and places using Google Maps.

    This requires an API key from Google with the following APIs enabled: -
    Geocoding API - Places API - Distance Matrix API

    Args:
        api_key: Google Maps API key default_location: Optional default
            location to use if none provided

        formatted_str: Whether to return a formatted string or a dictionary

        radius_km: Default search radius in kilometers (default: 10)

        force_distance: If set, this will force the distance to be included in
            the results

        enforced_limit: If set, this will override the limit argument
    """

    def __init__(
        self,
        api_key: str,
        default_location: Optional[str] = None,
        formatted_str: bool = False,
        radius_km: int = 10,
        force_distance: bool = False,
        enforced_limit: Optional[int] = None,
    ):
        self.__client = googlemaps.Client(key=api_key)
        self.__default_location = default_location
        self.__force_distance = force_distance
        self.__formatted_str = formatted_str
        self.__radius_km = radius_km

        args = [
            Argument(
                name="query",
                description=(
                    "What to search for (e.g., 'coffee shops', 'parks')",
                ),
                type="str",
                required=True,
            ),
        ]

        if not self.__default_location:
            args.append(
                Argument(
                    name="location",
                    description=(
                        "The location to search in (e.g., 'Seattle, WA')"
                    ),
                    type="str",
                    required=True,
                )
            )

        self.__enforced_limit = enforced_limit
        if not enforced_limit:
            args.append(
                Argument(
                    name="limit",
                    description="Maximum number of results to return",
                    type="int",
                    required=False,
                    default=enforced_limit,
                )
            )

        if not self.__force_distance:
            args.append(
                Argument(
                    name="radius",
                    description="Search radius in kilometers",
                    type="int",
                    required=False,
                    default=radius_km,
                )
            )

        examples = [
            Example(
                name="local_search",
                args={
                    "query": "ice cream",
                    "location": "Lafayette Square, St Louis",
                    "radius": 5,
                },
                output=(
                    "Search results near Lafayette Square, St Louis:\n\n"
                    "1. Jeni's Splendid Ice Creams\n"
                    "   • Distance: 2.1 km\n"
                    "   • Address: 389 N Euclid Ave, St. Louis, MO 63108, USA\n"
                    "   • Phone: (314) 367-1700\n"
                    "   • Website: https://jenis.com/scoop-shops/jenis-central-west-end/"
                    "2. Clementine's Naughty & Nice Ice Cream\n"
                    "   • Distance: 0.3 km\n"
                    "   • Address: 1637 S 18th St, St. Louis, MO 63104, USA\n"
                    "   • Phone: (314) 474-5800\n"
                    "   • Website: https://www.clementinescreamery.com/pages/lafayette-square"
                ),
                description="Search for ice cream shops near Lafayette Square",
            ),
        ]

        if self.__formatted_str:
            result = Result(
                type="str",
                description=(
                    "A formatted string containing local search results with "
                    "business details including distance"
                ),
            )
        else:
            result = Result(
                type="dict",
                description=(
                    "A dictionary containing search results. Format: "
                    "{'results': [{'name': str, 'address': str, 'phone': str, "
                    "'url': str, 'distance_km': float}]}"
                ),
            )

        super().__init__(
            name="LocalSearch",
            description="Search for local businesses and places using maps data",
            args=args,
            func=self.search,
            examples=examples,
            result=result,
        )

    def _format_results(self, results: List[dict], location: str) -> str:
        """Format results as a human-readable string."""
        if not results:
            return "No results found."

        output = f"Search results near {location}:\n\n"

        for i, result in enumerate(results, 1):
            output += f"{i}. {result['name']}\n"
            output += f"   • Distance: {result['distance_km']} km\n"

            if result["address"]:
                output += f"   • Address: {result['address']}\n"

            if result["phone"]:
                output += f"   • Phone: {result['phone']}\n"

            if result["url"]:
                output += f"   • Website: {result['url']}\n"

            output += "\n"

        return output.strip()

    def _calculate_distance(
        self,
        origin_coords: tuple[float, float],
        destination_coords: tuple[float, float],
    ) -> float:
        """
        Calculate the driving distance between two coordinate pairs.

        Args:
            origin_coords: Tuple of (latitude, longitude) for starting point
            destination_coords: Tuple of (latitude, longitude) for destination

        Returns:
            Distance in kilometers
        """
        distance = self.__client.distance_matrix(
            origins=f"{origin_coords[0]},{origin_coords[1]}",
            destinations=f"{destination_coords[0]},{destination_coords[1]}",
            mode="driving",
        )

        # Extract distance in meters from the response
        distance_meters = distance["rows"][0]["elements"][0]["distance"][
            "value"
        ]

        # Convert to kilometers and round to 1 decimal place
        return round(distance_meters / 1000, 1)

    def _normalize_results(
        self,
        raw_results: List[dict],
        detailed_results: List[dict],
        origin_coords: tuple,
        limit: int,
    ) -> List[dict]:
        """Convert Google Places results to our standard format."""
        results = []
        for raw, detailed in zip(raw_results[:limit], detailed_results):
            # Get place coordinates
            place_lat = raw["geometry"]["location"]["lat"]
            place_lng = raw["geometry"]["location"]["lng"]

            # Calculate distance using helper method
            distance_km = self._calculate_distance(
                origin_coords, (place_lat, place_lng)
            )

            result = {
                "name": detailed["name"],
                "address": detailed.get("formatted_address", ""),
                "phone": detailed.get("formatted_phone_number", ""),
                "url": detailed.get("website", ""),
                "distance_km": distance_km,
            }
            results.append(result)
        return results

    def search(
        self,
        query: str,
        location: Optional[str] = None,
        radius: Optional[int] = None,
        limit: Optional[int] = 5,
    ) -> Union[str, dict]:
        """
        Search for local businesses and places.

        Args:
            query: What to search for
            location: Where to search (if no default set)
            radius: Search radius in kilometers
            limit: Maximum number of results to return

        Returns:
            Either a formatted string or dictionary of results
        """
        location = location or self.__default_location
        if not location:
            raise ValueError("No location provided and no default location set")

        if self.__enforced_limit:
            limit = self.__enforced_limit

        if self.__force_distance:
            radius = self.__radius_km
        else:
            radius = radius or self.__radius_km

        try:
            # First, geocode the location to get coordinates
            geocode_result = self.__client.geocode(location)
            if not geocode_result:
                raise ValueError(f"Could not find location: {location}")

            lat = geocode_result[0]["geometry"]["location"]["lat"]
            lng = geocode_result[0]["geometry"]["location"]["lng"]

            # Store the origin coordinates for distance calculation
            origin_coords = (lat, lng)

            # Perform the places search
            places_result = self.__client.places_nearby(
                location=(lat, lng),
                radius=radius * 1000,  # Convert km to meters
                keyword=query,
            )

            if not places_result.get("results"):
                if self.__formatted_str:
                    return "No results found."
                return {"results": []}

            # Get detailed information for each place
            detailed_results = []
            raw_results = places_result["results"][:limit]
            for place in raw_results:
                place_details = self.__client.place(
                    place["place_id"],
                    fields=[
                        "name",
                        "formatted_address",
                        "formatted_phone_number",
                        "website",
                    ],
                )
                if place_details.get("result"):
                    detailed_results.append(place_details["result"])

            results = self._normalize_results(
                raw_results, detailed_results, origin_coords, limit
            )

            if self.__formatted_str:
                return self._format_results(results, location)
            return {"results": results}

        except Exception as e:
            raise Exception(f"Failed to perform local search: {str(e)}")



from __future__ import annotations

import email
import hashlib
import imaplib
import json
import os
import re
import traceback
from abc import ABC, abstractmethod
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime, timedelta
from threading import Lock, Thread
from time import sleep
from typing import Callable, Dict, List, Optional, Union

from arkaine.tools import Tool
from arkaine.utils.interval import Interval


class EmailMessage:
    def __init__(
        self,
        subject: str,
        sender: str,
        body: str,
        received_at: datetime,
        message_id: str,
        tags: List[str] = [],
    ):
        self.subject = subject
        self.sender = sender
        self.body = body
        self.received_at = received_at
        self.message_id = message_id
        self.tags = tags

        if not self.message_id:
            self.__generate_id()

    def __str__(self):
        """Convert EmailMessage to string representation"""
        # Remove microseconds from datetime representation
        formatted_date = self.received_at.replace(microsecond=0).isoformat()
        return (
            f"ID: {self.message_id}\n"
            f"From: {self.sender} @ {formatted_date}\n"
            f"Subject: {self.subject}\n"
            f"Date: {formatted_date}\n"
            f"Tags: {', '.join(self.tags)}\n"
            f"Body: {self.body}"
        )

    def __generate_id(self):
        """
        If the e-mail is lacking an ID, generate one
        from the combination of sender, subject, and body
        """
        if not self.message_id:
            self.message_id = hashlib.sha256(
                f"{self.sender}{self.subject}{self.body}".encode()
            ).hexdigest()

    @staticmethod
    def to_json(message: EmailMessage) -> str:
        """Convert EmailMessage to JSON string"""
        data = {
            "subject": message.subject,
            "sender": message.sender,
            "body": message.body,
            "received_at": message.received_at.isoformat(),
            "message_id": message.message_id,
            "tags": message.tags,
        }
        return json.dumps(data)

    @staticmethod
    def from_json(json_str: str) -> EmailMessage:
        """Create EmailMessage from JSON string"""
        data = json.loads(json_str)
        # Convert ISO format string back to datetime
        data["received_at"] = datetime.fromisoformat(data["received_at"])
        return EmailMessage(**data)

    @staticmethod
    def from_str(email_str: str) -> EmailMessage:
        """Create EmailMessage from string representation"""
        # Extract fields from string format
        lines = email_str.strip().split("\n")
        message_id = lines[0].replace("ID: ", "")
        sender = lines[1].split(" @ ")[0].replace("From: ", "")
        # Parse datetime without microseconds
        received_at = datetime.fromisoformat(lines[1].split(" @ ")[1])
        subject = lines[2].replace("Subject: ", "")
        tags_str = lines[4].replace("Tags: ", "")
        tags = tags_str.split(", ") if tags_str else []
        body = lines[5].replace("Body: ", "")

        return EmailMessage(
            subject=subject,
            sender=sender,
            body=body,
            received_at=received_at,
            message_id=message_id,
            tags=tags,
        )

    @staticmethod
    def from_message(msg: email.message.Message) -> EmailMessage:
        """Create EmailMessage from email.message.Message object."""
        subject = msg.get("subject", "")
        sender = msg.get("from", "")
        received_at = email.utils.parsedate_to_datetime(msg.get("date"))
        message_id = msg.get("message-id", "")

        # Decode the subject properly
        subject = email.header.decode_header(subject)
        subject = "".join(
            (
                str(t[0], t[1] if t[1] else "utf-8")
                if isinstance(t[0], bytes)
                else t[0]
            )
            for t in subject
        )

        body = ""
        if msg.is_multipart():
            parts = msg.walk()
        else:
            parts = [msg]

        for part in parts:
            if part.get_content_type() in ["text/plain", "text/html"]:
                try:
                    body = part.get_payload(decode=True).decode()
                except UnicodeDecodeError:
                    body = part.get_payload(decode=True).decode("ISO-8859-1")
                break

        return EmailMessage(
            subject=subject,
            sender=sender,
            body=body,
            received_at=received_at,
            message_id=message_id,
        )


class SeenMessageStore(ABC):
    def __init__(self):
        super().__init__()

    @abstractmethod
    def add(self, message: EmailMessage):
        pass

    @abstractmethod
    def contains(self, message: Union[EmailMessage, str]) -> bool:
        pass


class FileSeenMessageStore(SeenMessageStore):
    def __init__(self, filename: str):
        super().__init__()
        self.filename = filename
        self.__messages = set()
        self.__lock = Lock()

        if not os.path.exists(self.filename):
            self.save()
        else:
            self.load()

    def add(self, message: EmailMessage):
        with self.__lock:
            self.__messages.add(message.message_id)
            self.save()

    def contains(self, message: Union[EmailMessage, str]) -> bool:
        with self.__lock:
            if isinstance(message, EmailMessage):
                return message.message_id in self.__messages
            else:
                return message in self.__messages

    def save(self):
        with open(self.filename, "w") as f:
            f.write("\n".join(self.__messages))

    def load(self):
        with open(self.filename, "r") as f:
            self.__messages = set(f.readlines())


class EmailFilter:
    """
    A filter for EmailMessage objects that allows filtering based on various
    fields.

    Args:
        subject_pattern - (Optional[str]): A string or regex pattern to match
            the subject of the email.

        sender_pattern - (Optional[str]): A string or regex pattern to match
            the sender of the email.

        body_pattern -  (Optional[str]): A string or regex pattern to match the
            body of the email.

        tags -  (Optional[List[str]]): A list of tags to match against the
            email's tags.

        func -  (Optional[Callable[[EmailMessage], bool]]): A custom function
            to apply additional filtering logic.

        match_all -  (bool): If True, all conditions must be met; if False, ANY
            condition can be met and the filter will return True. Defaults to
            True.


    Note that filters can be combined by adding them, resulting in a set of
    filters that check in the order they are created. Thus if you have filters
    A and B, and you add them together in the order of A + B, then the
    resulting new filter will check A first, then B, and if both of the filters
    return True, then the new filter will return True.
    """

    def __init__(
        self,
        subject_pattern: Optional[str] = None,
        sender_pattern: Optional[str] = None,
        body_pattern: Optional[str] = None,
        tags: Optional[List[str]] = None,
        func: Optional[Callable[[EmailMessage], bool]] = None,
        match_all: bool = True,
    ):
        # Validate and compile regex patterns immediately
        self.subject_pattern = None
        self.sender_pattern = None
        self.body_pattern = None

        if subject_pattern:
            try:
                self.subject_pattern = re.compile(subject_pattern)
            except re.error as e:
                raise re.error(f"Invalid subject pattern: {e}")

        if sender_pattern:
            try:
                self.sender_pattern = re.compile(sender_pattern)
            except re.error as e:
                raise re.error(f"Invalid sender pattern: {e}")

        if body_pattern:
            try:
                self.body_pattern = re.compile(body_pattern)
            except re.error as e:
                raise re.error(f"Invalid body pattern: {e}")

        self.tags = [t.lower() for t in tags or []]
        self.func = func
        self.match_all = match_all

    def __call__(self, message: EmailMessage) -> bool:
        checks = []

        # Check subject pattern
        if self.subject_pattern is not None:
            if message.subject is None:
                checks.append(False)
            else:
                checks.append(
                    bool(self.subject_pattern.search(message.subject))
                )

        # Check sender pattern
        if self.sender_pattern is not None:
            if message.sender is None:
                checks.append(False)
            else:
                checks.append(bool(self.sender_pattern.search(message.sender)))

        # Check body pattern
        if self.body_pattern is not None:
            if message.body is None:
                checks.append(False)
            else:
                checks.append(bool(self.body_pattern.search(message.body)))

        # Check tags
        if self.tags:
            checks.append(any(tag.lower() in message.tags for tag in self.tags))

        # Check custom function
        if self.func is not None:
            try:
                checks.append(bool(self.func(message)))
            except Exception:
                checks.append(False)

        if not checks:
            return True

        return all(checks) if self.match_all else any(checks)

    def __add__(self, other):
        if not isinstance(other, (EmailFilter, Callable)):
            raise TypeError(f"Cannot combine EmailFilter with {type(other)}")
        return CombinedEmailFilter([self, other])


class CombinedEmailFilter:
    def __init__(
        self, filters: List[Union[EmailFilter, Callable[[EmailMessage], bool]]]
    ):
        self.__filters = []
        for f in filters:
            if isinstance(f, EmailFilter):
                self.__filters.append(f)
            elif isinstance(f, Callable):
                self.__filters.append(EmailFilter(func=f))
            else:
                raise TypeError(f"Invalid filter type: {type(f)}")

    def __call__(self, message: EmailMessage) -> bool:
        return all(f(message) for f in self.__filters)

    def __add__(self, other):
        if not isinstance(other, (EmailFilter, Callable, CombinedEmailFilter)):
            raise TypeError(f"Cannot combine with {type(other)}")

        new_filters = self.__filters.copy()
        if isinstance(other, CombinedEmailFilter):
            new_filters.extend(other.__filters)
        elif isinstance(other, EmailFilter):
            new_filters.append(other)
        elif isinstance(other, Callable):
            new_filters.append(EmailFilter(func=other))

        return CombinedEmailFilter(new_filters)


class Inbox:
    """
    Monitors incoming emails and triggers tools when matching emails are found.

    Basic usage:
    ```python
    # Create monitor checking every 5 minutes
    inbox = Inbox(
        username="user@example.com",
        password="pass",
        service="gmail",
        check_every="5:minutes"
    )

    # Add filter and associated tools
    inbox.add_filter(
        EmailFilter(subject_pattern="Important:.*"),
        tools=[notification_tool]
    )

    # Start monitoring
    inbox.start()

    # Add another filter while running
    inbox.add_filter(
        EmailFilter(sender_pattern="boss@company.com"),
        tools=[urgent_tool]
    )

    # Stop monitoring
    inbox.stop()
    ```
    """

    # Common IMAP servers
    COMMON_IMAP_SERVERS = {
        "gmail": "imap.gmail.com",
        "outlook": "outlook.office365.com",
        "yahoo": "imap.mail.yahoo.com",
        "aol": "imap.aol.com",
        "icloud": "imap.mail.me.com",
    }

    def __init__(
        self,
        call_when: Dict[
            Union[List, EmailFilter, Callable[[EmailMessage], bool]],
            Union[Tool, List[Tool]],
        ],
        username: Optional[Union[str, dict]] = None,
        password: Optional[Union[str, dict]] = None,
        service: Optional[str] = None,
        imap_host: Optional[str] = None,
        check_every: Union[str, Interval] = "5:minutes",
        env_prefix: str = "EMAIL",
        folders: Optional[List[str]] = None,
        store: Optional[SeenMessageStore] = None,
        ignore_emails_older_than: Optional[datetime] = None,
        max_messages_to_process: Optional[int] = None,
        use_ssl: bool = True,
    ):
        self.username = self._load_credential(
            username, f"{env_prefix}_USERNAME"
        )
        self.password = self._load_credential(
            password, f"{env_prefix}_PASSWORD"
        )
        self.__use_ssl = use_ssl

        if store:
            self.__seen_msg_store = store
        else:
            file_location = f"{env_prefix}_SEEN_MESSAGES.json"
            print(
                f"No store provided, defaulting to file store {file_location}"
            )
            self.__seen_msg_store = FileSeenMessageStore(file_location)

        # Set up IMAP host
        if imap_host:
            self.imap_host = imap_host
        elif service:
            if service not in self.COMMON_IMAP_SERVERS:
                raise ValueError(f"Unknown email service: {service}")
            self.imap_host = self.COMMON_IMAP_SERVERS[service]
        else:
            self.imap_host = os.getenv(f"{env_prefix}_IMAP_HOST")
            if not self.imap_host:
                raise ValueError("IMAP host not provided")

        # Set up interval
        if isinstance(check_every, str):
            self.interval = Interval(datetime.now(), recur_every=check_every)
        else:
            self.interval = check_every

        self.__tools: Dict[str, Tool] = {}
        self.__call_when: Dict[EmailFilter, List[str]] = {}

        for filters, tools in call_when.items():
            # Convert everything to a singular EmailFilter
            if isinstance(filters, List):
                filter = None
                for filter in filters:
                    if isinstance(filter, EmailFilter):
                        if filter:
                            filter += filter
                        else:
                            filter = filter
                    elif callable(filter):
                        filter += EmailFilter(func=filter)
                    else:
                        raise ValueError(f"Unknown filter type: {type(filter)}")
            elif isinstance(filters, EmailFilter):
                filter = filters
            elif callable(filters):
                filter = EmailFilter(func=filters)
            else:
                raise ValueError(f"Unknown filter type: {type(filters)}")

            if isinstance(tools, Tool):
                tools = [tools]

            self.__call_when[filter] = [t.tname for t in tools]
            for tool in tools:
                self.__tools[tool.tname] = tool

        self.__folders: List[str] = []
        if folders:
            self.__folders = folders
        else:
            self.__folders = ["INBOX"]

        self.__listeners: Dict[str, List[Callable]] = {
            "send": [],
            "error": [],
        }
        self.__threadpool = ThreadPoolExecutor()

        self.__ignore_emails_older_than = ignore_emails_older_than
        if self.__ignore_emails_older_than is None:
            # We default to the last 48 hours if not specified
            # to prevent people accidentally pulling years of emails.
            self.__ignore_emails_older_than = datetime.now() - timedelta(
                hours=48
            )

        self.__max_messages_to_process = max_messages_to_process

        self.__running = False
        self.__lock = Lock()
        self.__thread = Thread(target=self.__run)
        self.__thread.daemon = True
        self.__thread.start()

    def _load_credential(
        self,
        value: Optional[Union[str, dict]],
        env_var: str,
        required: bool = True,
    ) -> Optional[str]:
        """Load a credential from various sources."""
        if isinstance(value, str):
            return value
        if isinstance(value, dict) and "env" in value:
            value = os.getenv(value["env"])
        else:
            value = os.getenv(env_var)

        if required and not value:
            raise ValueError(
                f"Required credential not provided: direct value, "
                f"dict with 'env', or environment variable {env_var}"
            )
        return value

    def add_listener(self, on: str, listener: Callable):
        """
        Add a listener to the inbox, either triggering
        when an email is sent or when an error occurs.

        on can either be "send" or "error".

        If on is "send", then the listener will follow the
        type of func(EmailMessage, Filter, Context).

        If the on is "error", then the listener will follow the type of
        func(Exception).
        """
        with self.__lock:
            if on not in self.__listeners:
                raise ValueError(f"Unknown listener type: {on}")
            self.__listeners[on].append(listener)

    def start(self):
        """Start monitoring emails."""
        with self.__lock:
            self.__running = True

    def stop(self):
        """Stop monitoring emails."""
        with self.__lock:
            self.__running = False

    def __del__(self):
        if hasattr(self, "_Inbox__lock"):
            self.stop()
        if hasattr(self, "_Inbox__thread"):
            self.__thread.join()
        if hasattr(self, "_Inbox__threadpool"):
            self.__threadpool.shutdown(wait=True)

    def _check_emails(self):
        """Check for new emails and process them."""
        try:
            # Split host and port if provided in format "host:port"
            if ":" in self.imap_host:
                host, port = self.imap_host.split(":")
                port = int(port)
            else:
                host = self.imap_host
                port = 993 if self.__use_ssl else 143

            imap_class = imaplib.IMAP4_SSL if self.__use_ssl else imaplib.IMAP4
            with imap_class(host, port) as imap:
                imap.login(self.username, self.password)

                messages: List[EmailMessage] = []
                for folder in self.__folders:
                    try:
                        imap.select(folder)
                    except imaplib.IMAP4.error as e:
                        print(f"Error selecting folder {folder}: {e}")
                        continue

                    filter = ""
                    if self.__ignore_emails_older_than:
                        date_str = self.__ignore_emails_older_than.strftime(
                            "%d-%b-%Y"
                        )
                        filter += f'SINCE "{date_str}"'
                    if self.__max_messages_to_process:
                        filter += f"1:{self.__max_messages_to_process}"
                    if filter == "":
                        filter = "ALL"
                    _, message_numbers = imap.search(None, filter)

                    for num in message_numbers[0].split():
                        _, msg_data = imap.fetch(num, "(RFC822)")
                        email_body = msg_data[0][1]
                        msg = email.message_from_bytes(email_body)
                        message = EmailMessage.from_message(msg)

                        # Skip if we've seen this message before
                        if self.__seen_msg_store.contains(message):
                            continue
                        else:
                            messages.append(message)

                for msg in messages:
                    self.__seen_msg_store.add(msg)

                    # Check against filters and trigger tools
                    with self.__lock:
                        for email_filter, tools in self.__call_when.items():
                            if email_filter(msg):
                                for name in tools:
                                    tool = self.__tools[name]
                                    # ctx = tool.async_call(msg)
                                    ctx = None
                                    tool(msg)

                                # Trigger listeners
                                for listener in self.__listeners["send"]:
                                    self.__trigger_listener(
                                        listener, msg, email_filter, ctx
                                    )

        except Exception as e:
            trace = traceback.format_exc()
            print(f"Error checking emails: {str(e)}")
            print(trace)

            # Trigger error listeners
            with self.__lock:
                for listener in self.__listeners["error"]:
                    self.__trigger_listener(listener, e)

    def __trigger_listener(self, func: callable, **args):
        """
        Async fire off the listener
        """
        self.__threadpool.submit(func, **args)

    def __run(self):
        """Main monitoring loop."""
        while True:
            sleep(0.1)
            with self.__lock:
                if not self.__running:
                    continue

                check = self.interval.trigger_at <= datetime.now()

            if check:
                self._check_emails()
                self.interval.trigger()



from __future__ import annotations

import json
import pathlib
from os import path
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

from pydantic import BaseModel

from arkaine.backends.backend import Backend, ToolNotFoundException
from arkaine.llms.llm import LLM, Prompt
from arkaine.tools.argument import InvalidArgumentException
from arkaine.tools.context import Context
from arkaine.tools.tool import Tool
from arkaine.tools.types import ToolArguments, ToolResults
from arkaine.utils.templater import PromptTemplate


class ReActResponse(BaseModel):
    Thought: str
    Action: Optional[str] = None
    ActionInput: Optional[Dict[str, Any]] = None
    Answer: Optional[str] = None


class ReActBackend(Backend):

    def __init__(
        self,
        llm: LLM,
        tools: List[Tool],
        agent_explanation: str,
        initial_state: Dict[str, Any] = {},
        process_answer: Optional[Callable[[Any], Any]] = None,
        ignore_actions_without_input: bool = True,
    ):
        super().__init__(
            llm,
            tools,
            max_simultaneous_tools=1,
            initial_state=initial_state,
            process_answer=process_answer,
        )

        self.agent_explanation = agent_explanation
        self.__templater = PromptTemplate.from_file(
            path.join(
                pathlib.Path(__file__).parent,
                "prompts",
                "react.prompt",
            )
        )
        self.ignore_actions_without_input = ignore_actions_without_input

    def __parse(self, text: str) -> ReActResponse:
        lines = text.strip().split("\n")
        results: Dict[str, Optional[Union[str, Dict]]] = {
            "Thought": None,
            "Action": None,
            "Action Input": None,
            "Answer": None,
        }

        # Extract Thought
        if lines and lines[0].startswith("Thought:"):
            results["Thought"] = lines.pop(0).split("Thought:", 1)[1].strip()
        else:
            # raise FormatException
            results["Thought"] = ""

        # Extract Action and Action Input
        while lines:
            line = lines.pop(0)
            if not line.strip():
                continue
            if line.startswith("Action:"):
                results["Action"] = line.split("Action:", 1)[1].strip()
            elif line.startswith("Action Input:"):
                action_input_str = line.split("Action Input:", 1)[1].strip()
                try:
                    # First try to parse as JSON
                    results["Action Input"] = json.loads(action_input_str)
                except json.JSONDecodeError:
                    try:
                        # If JSON fails, try evaluating as Python literal
                        # Replace None, True, False with their JSON equivalents
                        action_input_str = (
                            action_input_str.replace("None", "null")
                            .replace("True", "true")
                            .replace("False", "false")
                        )
                        results["Action Input"] = json.loads(action_input_str)
                    except json.JSONDecodeError:
                        # If both fail, use the raw string
                        results["Action Input"] = action_input_str
            elif line.startswith("Answer:"):
                # Found the answer, capture it and any remaining lines
                results["Answer"] = (
                    line.split("Answer:", 1)[1].strip()
                    + "\n"
                    + "\n".join(lines)
                )
                break  # Stop processing after finding the answer

        # Validation
        if results["Action"] is not None and results["Action Input"] is None:
            if not self.ignore_actions_without_input:
                raise ValueError("Action specified without Action Input")
            else:
                results["Action"] = None
                results["Action Input"] = None

        # Handle missing Answer if Action is present - necessary for
        # pydantic
        if results["Action"] is not None and results["Answer"] is None:
            results["Answer"] = ""

        # Convert Action Input to ActionInput for pydantic
        results["ActionInput"] = results["Action Input"]
        del results["Action Input"]

        # If everything is blank, usually the model has output
        # the answer without the thought or Answer label
        if (
            not results["Thought"]
            and not results["Action"]
            and not results["ActionInput"]
            and not results["Answer"]
        ):
            results["Answer"] = text.strip()

        # Use Pydantic for final validation and parsing
        return ReActResponse(**results)

    def parse_for_result(self, context: Context, text: str) -> str:
        return self.__parse(text).Answer

    def parse_for_tool_calls(
        self, context: Context, text: str, stop_at_first_tool: bool = False
    ) -> List[Tuple[str, ToolArguments]]:
        response = self.__parse(text)

        return (
            []
            if not response.Action
            else [
                (
                    response.Action,
                    response.ActionInput,
                )
            ]
        )

    def tool_results_to_prompts(
        self, context: Context, prompt: Prompt, results: ToolResults
    ) -> List[Prompt]:
        for name, args, result in results:
            out = f"---\n{name}("

            first_tool = True
            for arg, value in args.items():
                if first_tool:
                    first_tool = False
                else:
                    out += ", "
                out += f"{arg}="
                if isinstance(value, str):
                    out += f'"{value}"'
                else:
                    out += f"{value}"
            out += ") "

            if isinstance(result, InvalidArgumentException):
                out += "encountered an error with the arguments passed"
                out += f"for this tool:\n{result}\n"
                out += "Remember the tool expects the following arguments:\n"
                out += (
                    "\n".join(str(arg) for arg in self.tools[name].args) + "\n"
                )
            elif isinstance(result, ToolNotFoundException):
                out += "\nNo such tool exists.\n"
                out += "Remember you have access to the following tools: "
                out += f"{','.join(self.tools.keys())}\n"
            else:
                out += f"returned:\n{result}\n"

            prompt.append(
                {
                    "role": "system",
                    "content": out,
                }
            )
        return prompt

    def prepare_prompt(self, context: Context, **kwargs) -> Prompt:
        # Create the tools block
        tools_block = ""
        for _, tool in self.tools.items():
            tools_block += f"{tool}\n"

        if len(self.tools) > 1:
            tool_names = f"{', '.join(self.tools.keys())}"
        else:
            tool_names = f"The tool {list(self.tools.keys())[0]}"

        return self.__templater.render(
            {
                # "agent_explanation": self.agent_explanation,
                "tools_block": tools_block,
                "tool_names": tool_names,
                "task": kwargs["task"],
            }
        )


class FormatException(Exception):
    pass


class ResponseException(Exception):
    pass


class ToolException(Exception):
    pass



from __future__ import annotations

import shutil
import tempfile
from typing import Any, Dict, List, Optional, Tuple, Union
from uuid import uuid4

from docker import DockerClient
from docker.models.containers import Container as dContainer


class Volume:
    """
    Volume represents a Docker volume that can be used to persist data
    across container runs. It encapsulates the properties and behaviors
    associated with a Docker volume.

    Attributes:
        name (str): A unique identifier for the volume. If not provided,
            a UUID will be generated.
        remote (str): The path inside the container where the volume will
            be mounted. Default is "/data".
        image (Optional[str]): The Docker image associated with the volume.
        persist_volume (bool): Indicates whether the volume should persist
            after the container is stopped. Default is False.
        read_only (bool): Indicates whether the volume should be mounted as
            read-only. Default is False.

    Methods:
        delete():
            Deletes the volume, even if it is marked as persistent. This method
            is intended for manual cleanup of the volume when persistence is no
            longer needed.
    """

    def __init__(
        self,
        name: Optional[str] = None,
        remote: str = "/data",
        image: Optional[str] = None,
        persist_volume: bool = False,
        read_only: bool = False,
    ):
        if name is None:
            self.__name = f"arkaine-{uuid4()}"
        else:
            self.__name = name
        self.__remote = remote
        self.__image = image
        self.__persist_volume = persist_volume
        self.__read_only = read_only

    @property
    def name(self) -> str:
        return self.__name

    @property
    def remote(self) -> str:
        return self.__remote

    @property
    def read_only(self) -> bool:
        return self.__read_only

    @property
    def image(self) -> str:
        return self.__image

    @property
    def persist_volume(self) -> bool:
        return self.__persist_volume

    def mount_args(self) -> Dict[str, str]:
        return {
            "type": "volume",
            "source": self.__name,
            "target": self.__remote,
            "read_only": self.__read_only,
        }

    def delete(self):
        """
        Deletes the volume, *even if persist volume is set*; this
        is meant as an "overwrite" method to manually clean up the
        volume after its persistence is unneeded.
        """
        self.__persist_volume = False
        del self

    def __del__(self):
        if self.__persist_volume:
            pass
        else:
            self.__client.volumes.get(self.__name).remove()

    def to_json(self) -> Dict[str, Any]:
        return {
            "name": self.__name,
            "remote": self.__remote,
            "image": self.__image,
            "persist_volume": self.__persist_volume,
            "read_only": self.__read_only,
        }

    @classmethod
    def from_json(cls, json: Dict[str, Any]) -> Volume:
        return cls(
            name=json["name"],
            remote=json["remote"],
            image=json["image"],
            persist_volume=json["persist_volume"],
            read_only=json["read_only"],
        )


class BindVolume:
    """
    BindVolume represents a bind mount in Docker, allowing a local directory
    to be mounted into a container. This is useful for sharing files between
    the host and the container.

    Attributes:
        local (Optional[str]): The local path on the host machine to be
            mounted. If not provided, a temporary directory will be created.
        remote (str): The path inside the container where the bind mount will
            be located. Default is "/data".
        name (Optional[str]): A unique identifier for the bind volume. If not
            provided, a UUID will be generated.
        read_only (bool): Indicates whether the bind mount should be read-only.
            Default is False.
    """

    def __init__(
        self,
        local: Optional[str] = None,
        remote: str = "/data",
        name: Optional[str] = None,
        read_only: bool = False,
    ):
        if name is None:
            self.__name = f"arkaine-{uuid4()}"
        else:
            self.__name = name
        if local is None:
            self.__local = tempfile.mkdtemp()
            self.__tmpdir = True
        else:
            self.__local = local
            self.__tmpdir = False
        self.__remote = remote
        self.__read_only = read_only

    @property
    def name(self) -> str:
        return self.__name

    @property
    def local(self) -> str:
        return self.__local

    @property
    def remote(self) -> str:
        return self.__remote

    @property
    def read_only(self) -> bool:
        return self.__read_only

    @property
    def image(self) -> str:
        return self.__image

    @property
    def persist_volume(self) -> bool:
        return self.__persist_volume

    def move_to(self, from_path: str, to_path: str):
        pass

    def mount_args(self) -> Dict[str, str]:
        return {
            "type": "bind",
            "source": self.__local,
            "target": self.__remote,
            "read_only": self.__read_only,
        }

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass

    def __del__(self):
        if self.__tmpdir:
            shutil.rmtree(self.__local)

    def to_json(self) -> Dict[str, Any]:
        return {
            "name": self.__name,
            "local": self.__local,
            "remote": self.__remote,
            "read_only": self.__read_only,
        }

    @classmethod
    def from_json(cls, json: Dict[str, Any]) -> BindVolume:
        return cls(
            local=json["local"],
            remote=json["remote"],
            name=json["name"],
            read_only=json["read_only"],
        )


class Container:
    """
    Container represents a Docker container that can be created, started,
    and managed. It encapsulates the properties and behaviors associated
    with a Docker container.

    Attributes:
        name (str): A unique identifier for the container.
        image (str): The Docker image to use for the container. Default is
                     "alpine:latest".
        args (Dict[str, Any]): Arguments to pass to the container at runtime.
        env (Dict[str, Any]): Environment variables to set in the container.
        volumes (List[Union[BindVolume, Volume]]): A list of volumes to mount
            in the container.
        ports (Dict[str, str]): Port mappings between the container and the
            host.
        entrypoint (Optional[str]): The command to run when the container
            starts.
        command (Optional[str]): The command to execute in the container.

    Methods:
        start():
            Starts the container with the specified configuration, running a
            command to keep it alive.

        run(command: Optional[str]) -> Tuple[str, str]:
            Executes a command in the running container and returns the output
            and error messages.

        bash(command: str) -> str:
            Executes a bash command in the container.

        stop():
            Stops and removes the container.
    """

    def __init__(
        self,
        name: str,
        image: str = "alpine:latest",
        args: Dict[str, Any] = {},
        env: Dict[str, Any] = {},
        volumes: List[Union[BindVolume, Volume]] = [],
        ports: Dict[str, str] = [],
        entrypoint: str = None,
        command: Optional[str] = None,
    ):

        self.__name = name
        self.__image = image
        self.__args = args
        self.__env = env
        self.__volumes = volumes

        port_bindings = {}
        if ports:
            for container_port, host_port in ports.items():
                if "/" not in container_port:
                    container_port = f"{container_port}/tcp"
                port_bindings[container_port] = host_port
        self.__ports = port_bindings

        self.__entrypoint = entrypoint
        if command is None:
            self.__command = "sleep infinity"
        else:
            self.__command = command

        self._container: Optional[dContainer] = None
        self.__client = DockerClient.from_env()

    def __image_check(self):
        # Check to see if we have the image; if not, attempt
        # to pull it
        try:
            self.__client.images.get(self.__image)
        except:  # noqa: E722
            try:
                self.__client.images.pull(self.__image)
            except:  # noqa: E722
                pass

    def __wait_for_container(self):
        self._container.reload()
        if self._container.status != "running":
            raise DockerExecutionException(
                f"Container failed to start. Status: {self._container.status}"
            )

    def start(self):
        """
        start runs the container with a "sleep infinity" command to keep
        it alive and running until it is told to stop.
        """
        if self._container:
            return

        self.__image_check()

        self._container = self.__client.containers.run(
            self.__image,
            name=self.__name,
            command="sleep infinity",
            detach=True,
            mounts=[v.mount_args() for v in self.__volumes],
            ports=self.__ports,
            environment=self.__env,
            entrypoint=self.__entrypoint,
        )

        self.__wait_for_container()

    def run(self, command: Optional[str]) -> Tuple[str, str]:
        self.__image_check()

        command = self.__command if command is None else command

        if not self._container:
            self.start()

        # Execute execute execute!
        result = self._container.exec_run(
            command,
            stderr=True,  # Enable stderr capture
            demux=True,  # Split stdout/stderr apart
        )

        # Capture output
        stdout, stderr = result.output
        stdout = stdout.decode("utf-8") if stdout else ""
        stderr = stderr.decode("utf-8") if stderr else ""

        # Check for errors
        if result.exit_code != 0:
            raise DockerExecutionException(stdout, stderr)

        return stdout, stderr

    def bash(self, command: str) -> str:
        # Make sure escape characters are handled for quotations
        command = command.replace("'", "\\'")
        return self.run(f"/bin/bash -c '{command}'")

    def stop(self):
        if self._container:
            self._container.remove(force=True)
            self._container = None

    @property
    def container(self) -> dContainer:
        return self._container

    def __enter__(self):
        self.run(None)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.cleanup()

    def __del__(self):
        self.__exit__(None, None, None)
        self.__client.close()


class DockerExecutionException(Exception):
    def __init__(self, message: str, stderr_or_trace: str = None):
        self.message = message
        self.stderr_or_trace = stderr_or_trace
        super().__init__(self.message)

    def __str__(self):
        if self.stderr_or_trace:
            return f"Python code execution failed:\n{self.message}\n\nError output:\n{self.stderr_or_trace}"
        return f"Python code execution failed: {self.message}"



from __future__ import annotations

import json
import os
from abc import ABC, abstractmethod
from datetime import datetime
from os.path import join
from queue import PriorityQueue
from threading import Lock, Thread
from time import sleep
from typing import List, Optional, Union
from uuid import uuid4

from arkaine.internal.registrar import Registrar
from arkaine.tools.context import Context
from arkaine.tools.tool import Tool
from arkaine.utils.interval import Interval
from arkaine.utils.timer import Timer


class Task:

    def __init__(
        self,
        tool: Tool,
        args: dict,
        trigger_at: Union[Interval, datetime],
        paused: bool = False,
        id: Optional[str] = None,
        history: List[float] = None,
        history_length: int = 100,
    ):
        self.__id = id or str(uuid4())
        self.__tool = tool
        self.__args = args
        self.__paused = paused
        self.__history: List[float] = history or []
        self.__history_length = history_length
        self.__lock = Lock()

        # Convert datetime to Interval if needed
        if isinstance(trigger_at, datetime):
            self.__interval = Interval(trigger_at)
        else:
            self.__interval = trigger_at

    @property
    def id(self) -> str:
        return self.__id

    @property
    def tool(self) -> Tool:
        return self.__tool

    @property
    def args(self) -> dict:
        return self.__args

    @property
    def interval(self) -> Interval:
        with self.__lock:
            return self.__interval

    @property
    def trigger_at(self) -> datetime:
        with self.__lock:
            return self.__interval.trigger_at

    @property
    def paused(self) -> bool:
        with self.__lock:
            return self.__paused

    @paused.setter
    def paused(self, value: bool):
        with self.__lock:
            self.__paused = value

    @property
    def last_triggered(self):
        with self.__lock:
            return self.__interval.last_triggered

    @property
    def next(self) -> datetime:
        """Get the next trigger time."""
        return self.__interval.trigger_at

    def trigger(self) -> datetime:
        with self.__lock:
            return self.__interval.trigger()

    def __call__(self, context: Optional[Context] = None):
        if self.paused:
            raise ValueError("Task is paused")

        with self.__lock:
            self.__interval.trigger()

            with Timer() as timer:
                result = self.tool(context=context, **self.args)

            self.__history.append(timer.elapsed)
            if len(self.__history) > self.__history_length:
                self.__history.pop(0)

        return result

    def to_json(self) -> dict:
        return {
            "id": self.id,
            "tool": self.tool.name,
            "args": self.args,
            "trigger_at": self.__interval.to_json(),
            "paused": self.paused,
            "history": self.__history,
            "history_length": self.__history_length,
        }

    @classmethod
    def from_json(cls, json_dict: dict) -> Task:
        interval = Interval.from_json(json_dict["trigger_at"])
        return cls(
            tool=Registrar.get_tool(json_dict["tool"]),
            args=json_dict["args"],
            trigger_at=interval,
            paused=json_dict["paused"],
            history=json_dict["history"],
            history_length=json_dict["history_length"],
        )


class TaskStore(ABC):

    @abstractmethod
    def save(self, tasks: Union[Task, List[Task]], overwrite: bool = False):
        pass

    @abstractmethod
    def load(self, task: Union[Task, str]) -> Task:
        pass

    @abstractmethod
    def load_all(self) -> List[Task]:
        pass


class FileScheduleStore(TaskStore):
    def __init__(self, dir_path: str):
        self.dirpath = dir_path
        os.makedirs(os.path.dirname(dir_path), exist_ok=True)

    def save(self, tasks: Union[Task, List[Task]], overwrite: bool = True):
        if isinstance(tasks, Task):
            tasks = [tasks]

        for task in tasks:
            filepath = join(self.dirpath, task.id)

            # if the file already exists, do not overwrite
            if os.path.exists(filepath) and not overwrite:
                continue

            with open(filepath, "w") as f:
                json.dump(task.to_json(), f)

    def load(self, task: Union[Task, str]) -> Task:
        if isinstance(task, Task):
            task = task.id

        with open(join(self.dirpath, task)) as f:
            return Task.from_json(json.load(f))

    def load_all(self) -> List[Task]:
        tasks = []
        for file in os.listdir(self.dirpath):
            tasks.append(self.load(file))
        return tasks


class Schedule:

    def __init__(
        self,
        tasks: Union[Task, List[Task], TaskStore] = [],
    ):
        self.__lock = Lock()
        self.tasks = []
        self.running = False

        self._schedule = PriorityQueue()
        self._last_check_time = 0

        self.schedule_store = None
        if isinstance(tasks, Task):
            self.add_task(tasks)
        elif isinstance(tasks, list):
            for task in tasks:
                self.add_task(task)
        elif isinstance(tasks, TaskStore):
            self.schedule_store = tasks

            for task in self.schedule_store.load_all():
                if task not in self.tasks:
                    self.tasks.append(task)
                    self._schedule.put((task.next(), task))

    def add_task(self, task: Task):
        next_time = task.next
        if self.schedule_store:
            self.schedule_store.save(task, overwrite=False)

        with self.__lock:
            if task in self.tasks:
                return

            self.tasks.append(task)
            self._schedule.put((next_time, task))

    def remove_task(self, task: Task):
        with self.__lock:
            self.tasks.remove(task)

    def run(self):
        with self.__lock:
            if self.running:
                return
            self.running = True

        Thread(target=self._check_tasks).start()

    def stop(self):
        with self.__lock:
            self.running = False

    def __run_task(self, task: Task):
        try:
            task()
        except Exception as e:
            print(f"Error running task: {e}")
        finally:
            if self.schedule_store:
                self.schedule_store.save(task)

    def _check_tasks(self):
        while True:
            with self.__lock:
                if not self.running:
                    break

            if self._schedule.empty():
                sleep(0.5)
                continue

            next_time, task = self._schedule.queue[0]
            now = datetime.now()
            if next_time > now:
                with self.__lock:
                    self._last_check_time = now
                sleep(0.5)
                continue
            else:
                next_time, task = self._schedule.get()
                task.trigger()
                if task.interval.recur_every:
                    self._schedule.put((task.trigger_at, task))
                    if self.schedule_store:
                        self.schedule_store.save(task)
                Thread(target=self.__run_task, args=(task,)).start()

    def save(self):
        if self.schedule_store:
            self.schedule_store.save(self)
        else:
            raise ValueError("No schedule store provided")



from __future__ import annotations

import atexit
import json
import shutil
import tempfile
from abc import ABC, abstractmethod
from concurrent.futures import Future, ThreadPoolExecutor, wait
from datetime import datetime
from hashlib import md5
from os.path import join
from pathlib import Path
from threading import Lock, Thread
from time import sleep
from typing import Any, Dict, List, Optional, Union

import feedparser
from feedparser.util import FeedParserDict

from arkaine.tools import Tool
from arkaine.utils.interval import Interval
from arkaine.utils.website import Website


class Item:
    def __init__(
        self,
        title: str,
        description: str,
        link: str,
        published: str,
        content: str,
    ):
        self.title = title
        self.description = description
        self.link = link
        self.published = published
        self.content = content

    @staticmethod
    def format(entry: Dict[str, Any]) -> Dict[str, Any]:
        return {
            "title": entry.get("title", ""),
            "description": entry.get("description", ""),
            "link": entry.get("link", ""),
            "published": entry.get("published", ""),
            "content": entry.get("content", entry.get("summary", "")),
        }

    @staticmethod
    def from_feedparser(entry: FeedParserDict) -> Item:
        return Item(
            title=entry.get("title", ""),
            description=entry.get("description", ""),
            link=entry.get("link", ""),
            published=entry.get("published", ""),
            content=entry.get("content", entry.get("summary", "")),
        )

    def __str__(self):
        return (
            f"{self.title}\n - {self.published}\n"
            f"\t{self.description}\n"
            f"\t{self.link}\n"
        )

    def get_website(self) -> Website:
        return Website(self.link)

    @property
    def md5(self) -> str:
        id = f"{self.title}-{self.link}-{self.published}"
        return md5(id.encode()).hexdigest()

    def to_json(self) -> Dict[str, Any]:
        return {
            "title": self.title,
            "description": self.description,
            "link": self.link,
            "published": self.published,
            "content": self.content,
        }

    @staticmethod
    def from_json(json: Dict[str, Any]) -> Item:
        return Item(
            title=json.get("title", ""),
            description=json.get("description", ""),
            link=json.get("link", ""),
            published=json.get("published", ""),
            content=json.get("content", ""),
        )


class Feed:

    def __init__(
        self,
        url: str,
        check_every: Union[str, Interval],
    ):
        self.__url = url
        if isinstance(check_every, str):
            if not self.__validate_recurrence(check_every):
                raise ValueError(
                    f"Invalid recurrence value: {check_every} - must be one of "
                    f"N{Interval.SECONDS}, N{Interval.MINUTES}, N{Interval.HOURS}"
                )
            self.__interval = Interval(datetime.now(), recur_every=check_every)
        elif isinstance(check_every, Interval):
            self.__interval = check_every
        else:
            raise ValueError(
                f"Invalid recurrence value: {check_every} - must be one of "
                f"N{Interval.SECONDS}, N{Interval.MINUTES}, N{Interval.HOURS}"
            )
        self.__last_check: Optional[datetime] = None

    def __validate_recurrence(self, value: str) -> bool:
        return value.endswith(
            (Interval.SECONDS, Interval.MINUTES, Interval.HOURS)
        )

    @property
    def url(self):
        return self.__url

    @property
    def trigger_at(self):
        return self.__interval.trigger_at

    @property
    def last_check(self):
        return self.__last_check

    def __str__(self):
        return (
            f"Feed(url={self.__url}, interval={self.__interval}, "
            f"last_check={self.__last_check})"
        )

    def get(self) -> List[Item]:
        self.__last_check = datetime.now()

        try:
            feed = feedparser.parse(self.__url)
            return [Item.from_feedparser(entry) for entry in feed.entries]
        except Exception as e:
            print(f"Error parsing RSS feed {self.__url}: {e}")
            return []
        finally:
            self.__interval.trigger()

    def to_json(self) -> Dict[str, Any]:
        return {
            "url": self.__url,
            "interval": self.__interval.to_json(),
            "last_check": self.__last_check,
        }

    @staticmethod
    def from_json(json: Dict[str, Any]) -> Feed:
        feed = Feed(
            url=json.get("url", ""),
            check_every=json.get("interval", ""),
        )
        feed.__last_check = json.get("last_check", datetime.now())
        return feed


class Store(ABC):
    """Abstract base class for RSS feed stores."""

    @abstractmethod
    def save_feed(
        self,
        feed: Feed,
    ) -> None:
        """Add a new RSS feed route with associated tool and format handler."""
        pass

    @abstractmethod
    def load_feed(self, feed: Feed) -> Optional[Feed]:
        """Load an RSS feed route."""
        pass

    @abstractmethod
    def save_item(
        self,
        item: Item,
    ) -> None:
        """Add a new RSS feed route with associated tool and format handler."""
        pass

    @abstractmethod
    def load_item(self, item: Item) -> Optional[Item]:
        """Load an RSS feed route."""
        pass


class FileStore(Store):
    def __init__(self, dir: str):
        self.__dir = Path(dir)
        self.__dir.mkdir(parents=True, exist_ok=True)
        # Make feed dir
        self.__feed_dir = self.__dir / "feeds"
        self.__feed_dir.mkdir(parents=True, exist_ok=True)
        # Make items dir
        self.__items_dir = self.__dir / "items"
        self.__items_dir.mkdir(parents=True, exist_ok=True)

    def __feed_path(self, feed: Feed) -> Path:
        id = md5(feed.url.encode()).hexdigest()
        return join(self.__feed_dir, id)

    def __item_path(self, item: Item) -> Path:
        return join(self.__items_dir, item.md5)

    def save_feed(self, feed: Feed) -> None:
        with open(self.__feed_path(feed), "w") as f:
            json.dump(feed.to_json(), f)

    def load_feed(self, feed: Feed) -> Optional[Feed]:
        try:
            with open(self.__feed_path(feed), "r") as f:
                return Feed.from_json(json.load(f))
        except FileNotFoundError:
            return None

    def save_item(self, item: Item) -> None:
        with open(self.__item_path(item), "w") as f:
            json.dump(item.to_json(), f)

    def load_item(self, item: Item) -> Optional[Item]:
        try:
            with open(self.__item_path(item), "r") as f:
                return Item.from_json(json.load(f))
        except FileNotFoundError:
            return None


class TempFileStore(FileStore):
    """
    This is a file store that utilizes a temp directory, which cleans up
    on exit (assuming clean exit).
    """

    def __init__(self):
        self.__dir = tempfile.mkdtemp()
        atexit.register(shutil.rmtree, self.__dir)
        super().__init__(self.__dir)


class RSS:
    """
    RSS is a trigger class that monitors multiple RSS feeds and executes tools
    when new items are detected. When this occurs, a list of Items
    (arkaine.triggers.rss.Item) are passed to the associated tools.

    Basic usage:
    ```python
    # Create feeds with check intervals
    feeds = [
        # Check every 30 minutes
        Feed("http://example.com/rss", "30:minutes"),
        # Check every hour
        Feed("http://another.com/feed", "1:hours"),
        # Check every day
        Feed(
            "http://example.com/rss",
            Interval(datetime.now(), recur_every="daily"),
        ),
    ]

    # Create RSS trigger with feeds and tools
    rss = RSS(feeds=feeds, tools=[my_tool])

    # Start monitoring
    rss.start()

    # Add new feed while running
    rss.add_feed(Feed("http://new.com/rss", "15:minutes"))

    # Stop monitoring
    rss.stop()
    ```
    """

    def __init__(
        self,
        feeds: List[Feed],
        store: Optional[Store] = None,
        tools: Optional[Union[List[Tool], Tool]] = None,
        max_workers: int = 10,
        feed_timeout: int = 30,
    ):
        self.__feeds = feeds

        if store:
            self.__store = store
            for i, feed in enumerate(self.__feeds):
                self.__feeds[i] = self.__store.load_feed(feed)
        else:
            self.__store = TempFileStore()

        if not tools:
            tools = []
        elif isinstance(tools, Tool):
            tools = [tools]
        self.__tools = tools

        self.__running = False
        self.__die = False
        self.__lock = Lock()

        self.__thread = Thread(target=self.__run)
        self.__thread.start()

        self.__feed_threadpool = ThreadPoolExecutor(
            max_workers=max_workers, thread_name_prefix="rss_feed_pool"
        )
        self.__feed_timeout = feed_timeout

    def __del__(self):
        """Ensure proper cleanup of threads and executor"""
        self.__die = True
        if hasattr(self, "_RSS__thread") and self.__thread.is_alive():
            self.__thread.join(timeout=1.0)  # Wait up to 1 second
        if hasattr(self, "_RSS__feed_threadpool"):
            self.__feed_threadpool.shutdown(wait=False, cancel_futures=True)

    @property
    def feeds(self):
        with self.__lock:
            return self.__feeds

    def add_feed(self, feed: Feed):
        with self.__lock:
            self.__feeds.append(feed)
            if self.__store:
                self.__store.save_feed(feed)

    @property
    def tools(self):
        with self.__lock:
            return self.__tools

    def add_tool(self, tool: Tool):
        if any(t.id == tool.id for t in self.__tools):
            return
        with self.__lock:
            self.__tools.append(tool)

    def start(self):
        with self.__lock:
            self.__running = True

    def stop(self):
        """Stop RSS monitoring and cleanup resources"""
        with self.__lock:
            self.__running = False
            # Cancel any pending futures
            self.__feed_threadpool.shutdown(wait=False, cancel_futures=True)
            # Create new threadpool for potential restart
            self.__feed_threadpool = ThreadPoolExecutor(
                max_workers=self.__feed_threadpool._max_workers,
                thread_name_prefix="rss_feed_pool",
            )

    def __handle_feed(self, feed: Feed):
        if feed.trigger_at > datetime.now():
            return

        items = feed.get()

        # Here we filter out the items that we have already
        # saved and thus processed. Note that we don't track
        # success here yet - that's a TODO
        filtered_items = []
        for index, item in enumerate(items):
            if not self.__store.load_item(item):
                self.__store.save_item(item)
                filtered_items.append(item)

        if len(filtered_items) == 0:
            return

        with self.__lock:
            # Finally trigger all tools attached to the RSS trigger class
            # with the detected items
            for tool in self.__tools:
                tool.async_call(filtered_items)

    def __run(self):
        while True:
            futures: List[Future] = []

            with self.__lock:
                if self.__die:
                    break
                elif not self.__running:
                    continue

                for feed in self.__feeds:
                    future = self.__feed_threadpool.submit(
                        self.__handle_feed, feed
                    )
                    futures.append(future)

            wait(futures, timeout=self.__feed_timeout)

            sleep(1.0)



from typing import Any, Callable, Dict, List, Optional, Union

from arkaine.tools.argument import Argument
from arkaine.tools.tool import Context, Example, Tool
from arkaine.tools.toolify import toolify


class DoWhile(Tool):
    """
    A tool that executes a sequence of steps repeatedly while a condition is
    met. The condition is evaluated after each iteration, ensuring the steps are
    executed at least once.

    The pseudocode flow of do while is thus, assuming the tool is called Foo,
    and our do while wrapper is called dw:

    dw(context, arguments):
        continue = True
        while(continue):
            arguments = prepare_args(context, args, output)
            output = Foo(context, arguments)

            continue = condition(context, output

            if not continue:
                break

        if format_output is not None:
            return format_output(context, output)

        return output

    The context of the do while has the following state keys maintained within
    it during do while execution that may be useful to utilize in prepare_args
    or the condition function:

    context["x"]:
        - iteration: The current iteration number
        - args: A list of each set of arguments passed to the tool for each
            iteration, starting with the initial arguments the tool was called
            with.
        - outputs: A list of the outputs of the tool for each iteration

    Args:
        tool (Tool): The tool to repeatedly trigger

        stop_condition (Callable[[Context, Any], bool]): Function that evaluates
            whether to continue the loop. Takes the context and current output
            as arguments and returns a boolean.

        args (Optional[List[Argument]]): A list of arguments the dowhile tool
            accepts. Note that if not provided, the tool's arguments will be
            utilized. If you change the arguments, you *must* also use
            prepare_args to adjust the arguments for the wrapped tool
            appropriately.

        prepare_args (Callable[[Context, Dict[str, Any]], Dict[str, Any]]):
            Function that prepares the arguments for the tool call. Takes the
            context and current arguments and returns a dictionary of arguments.

        format_output (Optional[Callable[[Context, Any], Any]]): Function that
            formats the output of the tool. Takes the context and output and
            returns a formatted output.

        name Optional[(str)]: The name of the do-while tool; defaults to the
            tool's name w/ ":do_while" appended

        description Optional[(str)]: Description of what the do-while
            accomplishes and it's condition. If not provided defaults to the
            wrapped tool's description.

        examples (List[Example]): Example usage scenarios

        max_iterations (Optional[int]): Maximum number of iterations to prevent
            infinite loops. Defaults to None (unlimited).

    Note:
        If using functions instead of tools, ensure the context is passed and
        utilized correctly.
    """

    def __init__(
        self,
        tool: Union[Tool, Callable[[Context, Any], Any]],
        stop_condition: Callable[[Context, Any], bool],
        prepare_args: Optional[
            Callable[[Context, Dict[str, Any]], Dict[str, Any]]
        ],
        args: Optional[List[Argument]] = None,
        format_output: Optional[Callable[[Context, Any], Any]] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        examples: List[Example] = [],
        max_iterations: Optional[int] = None,
        id: Optional[str] = None,
    ):
        self.stop_condition = stop_condition
        self.max_iterations = max_iterations
        self.prepare_args = prepare_args
        self.format_output = format_output

        if isinstance(tool, Tool):
            self.tool = tool
        else:
            self.tool = toolify(tool)

        # Use default name if none provided
        if name is None:
            name = f"{self.tool.name}:do_while"

        # Use tool's description if none provided
        if description is None:
            description = (
                f"Repeatedly executes {self.tool.name} until condition is "
                f"met. Inherits arguments from {self.tool.name}."
            )

        if args is not None and len(args) > 0:
            if self.prepare_args is None:
                raise ValueError(
                    "prepare_args must be provided if args are provided"
                )
            args = args
        else:
            args = self.tool.args

        super().__init__(
            name=name,
            args=args,
            description=description,
            func=self._loop,
            examples=examples,
            id=id,
        )

    def _loop(self, context: Context, **kwargs) -> Any:
        args = kwargs

        context.init("args", [])
        context.init("outputs", [])

        while True:
            context.init("iteration", 0)
            context.increment("iteration", 1)

            if (
                self.max_iterations is not None
                and context["iteration"] > self.max_iterations
            ):
                raise ValueError("max iterations surpassed")

            if self.prepare_args is not None:
                args = self.prepare_args(context, args)

            context.append("args", args)

            output = self.tool(context, **args)

            context.append("outputs", output)

            if self.stop_condition(context, output):
                break

        if self.format_output is not None:
            return self.format_output(context, output)

        return output

    def retry(self, context: Context) -> Any:
        """
        Retry the tool call. This attempts to pick up where the do-while flow
        left off in the current iteration.
        """
        if context.attached is None:
            raise ValueError("no tool assigned to context")

        if context.attached != self:
            raise ValueError(
                f"context is not for {self.name}, is instead for "
                f"{context.attached.name}"
            )

        context.clear(executing=True)

        # Since context maintains state, we can just retry the tool call
        # withe the last known state and arguments
        return self._loop(context, **context["args"][-1])



from google.ai.generativelanguage import FunctionDeclaration

from arkaine.tools.tool import Tool


def python(
    tool: Tool,
    output_style: str = "standard",
    include_examples: bool = False,
    include_return: bool = True,
) -> str:
    """
    Generate a Python docstring for the given tool.

    Args:
        tool (Tool): The tool for which to generate the docstring.

        output_style (str): The style of output to request (e.g., "standard",
            "google", "numpy").

        include_examples (bool): Whether to include examples in the docstring.

        include_return (bool): Whether to include the return type in the
            docstring (if a Result is specified for the tool).

    Returns:
        str: The generated docstring.
    """
    # Build function signature with tool name and args
    func_sig = f"\ndef {tool.name}("
    if tool.args:
        func_sig += ", ".join(arg.name for arg in tool.args)
    func_sig += ")\n"

    docstring = func_sig + f'"""\n{tool.description}\n\n'

    # Add arguments based on the output style
    if tool.args:
        if output_style == "google":
            docstring += "Args:\n"
            for arg in tool.args:
                desc = arg.description if hasattr(arg, "description") else ""
                arg_desc = f"    {arg.name} ({arg.type}): {desc}\n"
                docstring += arg_desc
        elif output_style == "numpy":
            docstring += "Parameters\n----------\n"
            for arg in tool.args:
                desc = arg.description if hasattr(arg, "description") else ""
                arg_desc = f"{arg.name} : {arg.type}\n    {desc}\n"
                docstring += arg_desc
        elif output_style == "standard":
            docstring += "Args:\n"
            for arg in tool.args:
                desc = arg.description if hasattr(arg, "description") else ""
                arg_desc = f"    {arg.name} ({arg.type}): {desc}\n"
                docstring += arg_desc
        else:
            raise ValueError(f"Invalid output style: {output_style}")

    if tool.result and include_return:
        docstring += f"\n\nReturns: {tool.result}\n"

    if include_examples and tool.examples:
        docstring += "\nExamples:\n"
        for example in tool.examples:
            docstring += f"    {example}\n"

    docstring += '"""'
    return docstring


def openai(
    tool: Tool, include_examples: bool = False, include_return: bool = True
) -> str:
    """
    Generates a structured representation of a tool for OpenAI integration.

    This function takes a Tool object and constructs a dictionary that
    describes the tool's properties, including its name, description,
    parameters, and required arguments.

    Args:
        tool (Tool): The tool object containing metadata about the tool.

        include_examples (bool): Whether to include example usages in the
            output.

        include_return (bool): Whether to include return type information in
            the output.

    Returns:
        dict: A dictionary representation of the tool, suitable for OpenAI
            function calls.
    """
    properties = {}
    required_args = []

    description = tool.description
    if include_return and tool.result:
        description += f"\n\nReturns: {tool.result}"
    if include_examples and tool.examples:
        description += "\n\nExamples:\n"
        for example in tool.examples:
            description += f"    {example}\n"

    for arg in tool.args:
        arg_type = arg.type_str
        if arg_type == "str":
            arg_type = "string"

        properties[arg.name] = {
            "type": arg_type,
            "description": arg.description,
        }
        if arg.required:
            required_args.append(arg.name)

    return {
        "type": "function",
        "function": {
            "name": tool.name,
            "description": tool.description,
            "parameters": {
                "type": "object",
                "properties": properties,
                "required": required_args,
            },
        },
    }


def gemini(tool: Tool) -> FunctionDeclaration:
    """Convert a tool into a Gemini FunctionDeclaration.

    Args:
        tool: tool obj to convert

    Returns:
        FunctionDeclaration object compatible with Gemini's function calling
    """
    # Create the parameters schema
    parameters = {"type_": "OBJECT", "properties": {}, "required": []}

    # Convert each argument into the Gemini format
    for arg in tool.args:
        type_map = {
            "str": "STRING",
            "string": "STRING",
            "number": "NUMBER",
            "float": "NUMBER",
            "integer": "NUMBER",
            "int": "NUMBER",
            "boolean": "BOOLEAN",
            "array": "ARRAY",
            "object": "OBJECT",
            "dict": "OBJECT",
            "list": "ARRAY",
            "tuple": "ARRAY",
            "set": "ARRAY",
            "tuple": "ARRAY",
            "datetime": "STRING",
            "date": "STRING",
            "time": "STRING",
            "timedelta": "STRING",
        }

        # Get the Gemini type, defaulting to STRING if unknown
        gemini_type = type_map.get(arg.type_str.lower(), "STRING")

        description = arg.description
        if arg.default:
            description += f"\nDefault: {arg.default}"

        # Add the property
        parameters["properties"][arg.name] = {
            "type_": gemini_type,
            "description": (arg.description),
        }

        # Add to required list if necessary
        if arg.required:
            parameters["required"].append(arg.name)

    return FunctionDeclaration(
        name=tool.name, description=tool.description, parameters=parameters
    )



from setuptools import find_packages, setup
from typing import List


def read_requirements(filename: str) -> List[str]:
    """Read requirements from file, cleaning up formatting."""
    with open(filename) as f:
        requirements = []
        for line in f:
            line = line.strip()
            # Skip empty lines, comments, and section headers
            if not line or line.startswith("#") or line.startswith("["):
                continue
            # Keep version specifiers to ensure reproducible builds
            requirements.append(line)
        return requirements


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="arkaine",
    version="0.0.12",
    author="Keith Chester",
    author_email="k@hlfshell.ai",
    description="A batteries-included framework for DIY AI agents",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/hlfshell/arkaine",
    packages=find_packages(),
    install_requires=read_requirements("requirements.txt"),
    extras_require={
        "dev": [
            "pytest",
        ],
    },
    python_requires=">=3.8",  # Specify minimum Python version
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    entry_points={
        "console_scripts": [
            "spellbook=arkaine.spellbook.server:main",
        ],
    },
)



import pathlib
from os import path
from typing import Any, Callable, List, Optional

from arkaine.llms.llm import LLM, Prompt
from arkaine.tools.agent import Agent
from arkaine.tools.tool import Argument, Context, Example, Tool
from arkaine.utils.templater import PromptTemplate


class ArgsFormatter(Tool):
    """
    A tool wrapper that allows you to add or remove Arguments from tools,
    and/or alter the argument values being passed to a wrapped tool.

    Attributes:
        tool (Tool): The tool to be wrapped. kwargs_formatter
        (Callable[[Context, Any], Any]): A function to format the keyword
            arguments.
        add_args (Optional[List[Argument]]): Additional arguments to be added
            to the tool.
        remove_args (Optional[List[str]]): Arguments to be removed
            from the tool.
        name (Optional[str]): Custom name for the tool - otherwise it defaults
            to the tool's name
        description (Optional[str]): Custom description for the tool,
            defaulting to the tool's description if None is provided.
        examples (Optional[List[Example]]): Custom examples for the tool. If
            not provided, uses the wrapped tool's examples.
    """

    def __init__(
        self,
        tool: Tool,
        kwargs_formatter: Callable[[Context, Any], Any],
        add_args: Optional[List[Argument]] = None,
        remove_args: Optional[List[str]] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        examples: Optional[List[Example]] = None,
    ):
        self._tool = tool
        self._kwargs_formatter = kwargs_formatter
        self._name = name

        args = [arg for arg in tool.args]
        if add_args:
            args.extend(add_args)
        if remove_args:
            args = [arg for arg in args if arg.name not in remove_args]

        name = name or tool.name
        description = description or tool.description
        examples = examples or tool.examples

        super().__init__(name, description, args, self.format, examples)

    def format(self, context: Context, **kwargs: Any) -> Any:
        if self._kwargs_formatter:
            kwargs = self._kwargs_formatter(context, **kwargs)

        return self._tool()


class Formatter(Tool):
    """
    A tool wrapper that formats the output of a wrapped tool. This is useful
    if a tool developed by another developer and you find yourself needing to
    massage the output into a format more friendly to your chosen Agent or
    LLM.

    Attributes:
        tool (Tool): The tool to be wrapped.
        formatter (Callable[[Context, Any], Any]): A function to format the
            output of the tool. Takes the execution context and tool output as
            arguments and returns the formatted result.
        name (Optional[str]): Custom name for the tool. If not provided, uses
            the wrapped tool's name.
        description (Optional[str]): Custom description for the tool. If not
            provided, uses the wrapped tool's description.
        examples (Optional[List[Example]]): Custom examples for the tool. If
            not provided, uses the wrapped tool's examples.
    """

    def __init__(
        self,
        tool: Tool,
        formatter: Callable[[Context, Any], Any],
        name: Optional[str] = None,
        description: Optional[str] = None,
        examples: Optional[List[Example]] = None,
    ):
        self._tool = tool
        self._formatter = formatter

        name = name or tool.name
        description = description or tool.description
        examples = examples or tool.examples

        super().__init__(name, description, tool.args, self.format, examples)

    def format(self, context: Context, **kwargs: Any) -> Any:
        return self._formatter(context, **kwargs)


class FormattingAgent(Agent):
    """
    An agent that reformats the output of a tool using an LLM.

    Attributes:
        tool (Tool): The tool whose output needs to be reformatted.
        llm (LLM): TThe language model to use for reformatting.
        original_format_explanation (str): A description of the input format
            that the tool produces. This helps the LLM understand the structure
            of the data it's working with.
        formatted_format_explanation (str): A description of the desired output
            format that the agent should produce.
        name (Optional[str]): Custom name for the agent. Defaults to the tool's
            name if not provided.
        description (Optional[str]): Custom description for the agent. Defaults
            to the tool's description if not provided.
        examples (Optional[List[Example]]): Custom examples for the agent.
            Defaults to the tool's examples if not provided. tool's examples if
            not provided.
        process_answer (Optional[Callable[[str], Any]]): Optional function to
            process the LLM's response before returning it.
        prompt (Optional[PromptTemplate]): Custom prompt template. If not
            provided, a short default prompt is used.

    The prompt template should use the following placeholders:
        - {original_format}: Description of the input format
        - {formatted_format}: Description of the desired output format
        - {output}: The actual output from the tool that needs to be
          reformatted - this will be set on invocation.
    """

    def __init__(
        self,
        tool: Tool,
        llm: LLM,
        original_format_explanation: str,
        formatted_format_explanation: str,
        name: Optional[str] = None,
        description: Optional[str] = None,
        examples: Optional[List[Example]] = None,
        process_answer: Optional[Callable[[Context, str], Any]] = None,
        prompt: Optional[PromptTemplate] = None,
    ):
        self._tool = tool
        self._original_format_explanation = original_format_explanation
        self._formatted_format_explanation = formatted_format_explanation
        self._process_answer = process_answer

        name = name or tool.name
        description = description or tool.description
        examples = examples or tool.examples

        if prompt is None:
            self._prompt = PromptTemplate.from_file(
                path.join(
                    pathlib.Path(__file__).parent,
                    "prompts",
                    "formatting_agent.prompt",
                )
            )
        else:
            self._prompt = prompt

        super().__init__(name, description, tool.args, llm, examples)

    def invoke(self, context: Context, **kwargs) -> Any:
        output = self._tool(context=context, **kwargs)
        return super().invoke(context, output=output)

    def prepare_prompt(self, context: Context, **kwargs: Any) -> Prompt:
        return self._prompt.format(
            original_format=self._original_format_explanation,
            formatted_format=self._formatted_format_explanation,
            output=kwargs["output"],
        )

    def extract_result(self, context: Context, output: str) -> Any:
        if self._process_answer:
            return self._process_answer(context, output)
        return output
